/* -*- Mode: C++; tab-width: 20; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Anya client code
 *
 * The Initial Developer of the Original Code is Netscape Communications
 * Corporation. Portions created by the Initial Developer are
 * Copyright (C) 2003 the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *    Stuart Parmenter <pavlov@netscape.com>
 *    Joe Hewitt <hewitt@netscape.com>
 *    Ben Goodger <ben@netscape.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 */

#include "anya.h"
#include "client.h"
#include "oz.h"
#include "controls.h"
#include "utf8.h"

#ifdef _WIN32_WCE
#include <Aygshell.h>
#include <Commctrl.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define DOUBLE_BUFFER

#ifdef _WIN32_WCE
#define strdup(x) _strdup(x)
#endif

#ifndef IDC_HAND
#define IDC_HAND MAKEINTRESOURCE(32649)
#endif

/* XXX this should be more general.. allow passing in of a window, etc */
AnyaWindow *
anya_window_create(int x, int y, int width, int height)
{
    AnyaWindow *w;

    anya_init_controls();

    w = malloc(sizeof(AnyaWindow));
    if (!w)
        return NULL;

	/* take in to account the space the menubar wastes */
    w->window = oz_window_create(NULL, x, y, width, height);
    if (!w->window) {
        free(w);
        return NULL;
    }

    oz_window_set_owner(w->window, w);
    memset(&w->cd, 0, sizeof(ClientData));
    return w;
}

void
anya_window_destroy(AnyaWindow *win)
{
    oz_window_destroy(win->window);
    clientdata_destroy(&win->cd);
    free(win);
}


void
anya_window_load(AnyaWindow *win, const char *uri, uint8 method)
{
    /* XXX fix the scrollbar range code here */
#if 0
    int vsbw = GetSystemMetrics(SM_CXVSCROLL);
    int hsbh = GetSystemMetrics(SM_CYHSCROLL);
    RECT r;
    GetClientRect(win->window->hwnd, &r);
#endif
    oz_window_set_title(win->window, L"Loading...");
    clientdata_request(&win->cd, uri, method);
    
    anya_create_controls(win);
    oz_window_scroll_set_range(win->window,
                               0, win->cd.header.w /* - r.right */,
                               0, win->cd.header.h /* - r.bottom */);
    oz_window_scroll_set_pos(win->window, 0, 0);
    oz_window_redraw(win->window);
    oz_window_set_title(win->window, L"Loaded");
}

static int
pt_in_clip_rgn(HDC dc, uint16 x, uint16 y)
{
#ifdef _WIN32_WCE
	RECT r;
	r.left = r.right  = x;
	r.top  = r.bottom = y;
    return RectVisible(dc, &r);
#else
    return PtVisible(dc, x, y);
#endif
}

static int
rect_in_clip_rgn(HDC dc, uint16 x, uint16 y, uint16 w, uint16 h)
{
    RECT r;
    r.left   = x;
    r.top    = y;
    r.right  = x + w;
    r.bottom = y + h;
    return RectVisible(dc, &r);
}

static void
anya_window_draw(AnyaWindow *awin)
{
    const ozWindow *win = awin->window;
    ozGC *gc = &awin->gc;

    uint8 *curs = awin->cd.drawing;
    const uint8 *end = curs + awin->cd.header.display_list_size;

    PAINTSTRUCT paint;

#ifdef DOUBLE_BUFFER
    ozDoubleBuffer db;
#endif

    uint16 sx, sy;
    oz_window_scroll_get_pos(win, &sx, &sy);

    /* Start painting */
    gc->dc = BeginPaint(win->hwnd, &paint);

#ifdef DOUBLE_BUFFER
    oz_double_buffer_start(&db, gc,
                           paint.rcPaint.right - paint.rcPaint.left,
                           paint.rcPaint.bottom - paint.rcPaint.top);
    sx += paint.rcPaint.left;
    sy += paint.rcPaint.top;
#endif

    SetViewportOrgEx(gc->dc, -sx, -sy, NULL);

    while (curs < end) {
        switch (*curs) {
        case ABC_POINT:
            {
                const struct Point *point = (const struct Point *)curs;
                curs += sizeof(struct Point);
                if (!pt_in_clip_rgn(gc->dc, point->x, point->y))
                    break;
                oz_draw_point(win, gc, point->x, point->y);
            }
            break;
        case ABC_LINE:
            {
                const struct Line *line = (const struct Line *)curs;
                curs += sizeof(struct Line);
#if 0
                if (!rect_in_clip_rgn(gc->dc, line->x1, line->y1, line->x2 - line->x1, line->y2 - line->y1))
                    break;
#endif
                oz_draw_line(win, gc, line->x1, line->y1, line->x2, line->y2);
            }
            break;
        case ABC_RECT:
            {
                const struct Rect *rect = (const struct Rect *)curs;
                curs += sizeof(struct Rect);
                if (!rect_in_clip_rgn(gc->dc, rect->x, rect->y, rect->w, rect->h))
                    break;
                oz_draw_rect(win, gc, rect->x, rect->y, rect->w, rect->h, rect->filled);
            }
            break;
        case ABC_ARC:
            {
                const struct Arc *arc = (const struct Arc *)curs;
                curs += sizeof(struct Arc);
                oz_draw_arc(win, gc, arc->x1, arc->y1, arc->x2, arc->y2,
                            f9tof(arc->start), f9tof(arc->end));
            }
            break;
        case ABC_POINTS:
            {
                const struct Points *points = (const struct Points *)curs;
                curs += sizeof(struct Polygon);
                oz_draw_points(win, gc, points->points, points->npoints);
            }
            break;
        case ABC_POLYLINE:
            {
                const struct Polyline *polyline = (const struct Polyline *)curs;
                curs += sizeof(struct Polyline);
                oz_draw_lines(win, gc, polyline->points, polyline->npoints);
            }
            break;
        case ABC_POLYGON:
            {
                const struct Polygon *poly = (const struct Polygon *)curs;
                curs += sizeof(struct Polygon);
                oz_draw_polygon(win, gc, poly->points, poly->npoints, poly->filled);
            }
            break;
        case ABC_TEXT:
            {
                const struct Text *text = (const struct Text *)curs;
                uint32 stringlen;
                const char *string = idmap_GetString(&awin->cd.idmap, text->id, &stringlen);
                curs += sizeof(struct Text);
                /* XXX we should probably send over the width and height info from the server
                   for now use some numbers that should be slightly larger so that we
                   don't redraw ALL text on a page */
                if (!rect_in_clip_rgn(gc->dc, text->x, text->y, (uint16)(stringlen * 10), 25))
                    break;
                oz_draw_string(win, gc, text->x, text->y, string, stringlen);
            }
            break;
        case ABC_FONT:
            gc->font.id = ((const struct Font*)curs)->id;
            curs += sizeof(struct Font);
            break;
        case ABC_COLOR:
            {
                const struct Color *color = (const struct Color*)curs;
                gc->rgbaq = RGB(color->r,
                                color->g,
                                color->b);
                curs += sizeof(struct Color);
            }
            break;

        case ABC_IMAGE:
            {
                const struct Image *img = (const struct Image *)curs;
                curs += sizeof(struct Image);
                if (!rect_in_clip_rgn(gc->dc, img->x, img->y, img->destw, img->desth))
                    break;
                oz_draw_image(win, gc, img->x, img->y, img->w, img->h,
                              img->destw, img->desth, img->xoffset, img->yoffset,
                              img->istile, img->bpp, img->nbytes, img->bytes);
            }
            break;
        case ABC_INPUT_FORM:
            curs += sizeof(struct InputForm);
            break;
        case ABC_INPUT_HIDDEN:
            curs += sizeof(struct InputHidden);
            break;
        case ABC_INPUT_TEXT:
        case ABC_INPUT_PASSWORD:
        case ABC_INPUT_TEXTAREA:
            curs += sizeof(struct InputText);
            break;
        case ABC_INPUT_CHECKBOX:
            curs += sizeof(struct InputCheckbox);
            break;
        case ABC_INPUT_RADIO:
            curs += sizeof(struct InputRadio);
            break;
        case ABC_INPUT_SUBMIT:
            curs += sizeof(struct InputSubmit);
            break;
        case ABC_INPUT_SELECT:
            curs += sizeof(struct InputSelect);
            break;
        default:
            break;
        }
    }

    SetViewportOrgEx(gc->dc, 0, 0, NULL);

#ifdef DOUBLE_BUFFER
    oz_double_buffer_stop(&db, gc,
                          paint.rcPaint.left,
                          paint.rcPaint.top,
                          paint.rcPaint.right - paint.rcPaint.left,
                          paint.rcPaint.bottom - paint.rcPaint.top);
#endif

    /* End painting */
    EndPaint(win->hwnd, &paint);
    gc->dc = NULL;
}

static int
rect_contains(const ozRect *r, uint16 x, uint16 y)
{
    return ((x >= r->x) && (y >= r->y) &&
            (x < r->x + r->width) && (y < r->y + r->height));
}

static const struct Event *
look_for_event(uint16 x, uint16 y, const ClientData *cd)
{
    /* the list is stored backwards */
    uint8 *curs = cd->events + cd->event_list_size;
    const uint8 *end = cd->events;

    const struct Event *e;
    uint32 i;

    while (curs > end) {
        curs -= sizeof(struct Event);
        e = (const struct Event *)curs;
        for (i=0; i< e->nrects; ++i) {
            if (rect_contains(&e->rects[i], x, y))
                return e;
        }
    }

    return NULL;
}

static const struct Input *
look_for_input_element(uint16 x, uint16 y, const ClientData *cd)
{
    uint8 i;
    for (i = 0; i < cd->header.form_list_length; ++i) {
        const struct Input *input = cd->forms[i];
        switch(input->type) {
        case ABC_INPUT_TEXT:
        case ABC_INPUT_PASSWORD:
        case ABC_INPUT_TEXTAREA:
            {
                const struct InputText *inputText = (const struct InputText *)input;
                const ozRect r = { inputText->x, inputText->y, inputText->w, inputText->h };
                if (rect_contains(&r, x, y))
                    return input;
            }
            break;
        case ABC_INPUT_SUBMIT:
            {
                const struct InputSubmit *inputSubmit = (const struct InputSubmit *)input;
                const ozRect r = { inputSubmit->x, inputSubmit->y, inputSubmit->w, inputSubmit->h };
                if (rect_contains(&r, x, y))
                    return input;
            }
            break;
        default:
            break;
        }
    }

    return NULL;
}

static void
scroll_window(ozWindow *window, WPARAM wParam, int direction)
{
    long oldPos;
    SCROLLINFO si;
    si.cbSize = sizeof(si);
    si.fMask  = SIF_ALL;

    GetScrollInfo(window->hwnd, direction, &si);

    /* save this for later */
    oldPos = si.nPos;

    switch (LOWORD(wParam)) {
    case SB_LINELEFT:           /* user clicked left arrow */
        si.nPos -= 5;
        break;
    case SB_LINERIGHT:          /* user clicked right arrow */
        si.nPos += 5;
        break;
    case SB_PAGELEFT:           /* user clicked shaft left of the scroll box */
        si.nPos -= si.nPage;
        break;
    case SB_PAGERIGHT:          /* user clicked shaft right of the scroll box */
        si.nPos += si.nPage;
        break;
    case SB_THUMBTRACK:         /* user dragged the scroll box */
        si.nPos = si.nTrackPos;
        break;
    default :
        break;
    }

    /* Set the position and then retrieve it.  Due to adjustments
       by Windows it may not be the same as the value set. */
    si.fMask = SIF_POS;
    SetScrollInfo(window->hwnd, direction, &si, TRUE);
    GetScrollInfo(window->hwnd, direction, &si);

    /* If the position has changed, scroll the window  */
    if (si.nPos != oldPos) {
        int flags = SW_INVALIDATE | SW_SCROLLCHILDREN;
#ifdef _DEBUG
        /* Only use ERASE in debug builds.  We want the background color to
           get painted first incase nothing is drawn in the area */
        flags |= SW_ERASE;
#endif
        ScrollWindowEx(window->hwnd,
                       (direction == SB_HORZ) ? oldPos - si.nPos : 0,
                       (direction == SB_VERT) ? oldPos - si.nPos : 0,
                       NULL, NULL, NULL, NULL, flags);
        UpdateWindow(window->hwnd);
    }
}

LRESULT CALLBACK AnyaWndProc(HWND hwnd,
                             UINT message,
                             WPARAM wParam,
                             LPARAM lParam)
{
    AnyaWindow *awin;
#ifdef _WIN32_WCE
    static HWND menuBar;
    static SHACTIVATEINFO s_sai;
#endif
    awin = (AnyaWindow*)oz_window_get_owner(oz_window_from_hwnd(hwnd));

    switch (message) {
    case WM_COMMAND:
        {
            if (HIWORD(wParam) == BN_CLICKED) {
                HWND target = (HWND)lParam;
                DWORD style = GetWindowLong(target, GWL_STYLE);

                if (style & BS_RADIOBUTTON) {
                    struct InputRadio *input = (struct InputRadio *)oz_get_hwnd_userdata((HWND)lParam);
                    anya_handle_radio_click(awin, input);
                }
            }

            return 0;
        }
        break;
    case WM_MOUSEMOVE:
        {
            const struct Event *ev;
            uint16 sx, sy;
            oz_window_scroll_get_pos(awin->window, &sx, &sy);
            ev = look_for_event((uint16)(LOWORD(lParam) + sx),
                                (uint16)(HIWORD(lParam) + sy),
                                &awin->cd);
            SetCursor(LoadCursor(NULL, ev ? IDC_HAND : IDC_ARROW));

            return 0;
        }
        break;
    case WM_LBUTTONDOWN:
        {
            const struct Event *ev;
            const struct Input *in;
            char *uri = NULL;
            uint8 method = ANYA_REQUEST_GET;
            uint16 sx, sy;
            oz_window_scroll_get_pos(awin->window, &sx, &sy);

            if (ev = look_for_event((uint16)(LOWORD(lParam) + sx),
                                    (uint16)(HIWORD(lParam) + sy),
                                    &awin->cd)) {
                /* cliendata_destroy() will free the idmap, so we need to
                   copy the string here */
                uri = idmap_GetStringCopy(&awin->cd.idmap, ev->id, NULL);
                if (!uri) return 0;

            } else if (in = look_for_input_element((uint16)(LOWORD(lParam) + sx),
                                                   (uint16)(HIWORD(lParam) + sy),
                                                   &awin->cd)) {
                if (in->type == ABC_INPUT_SUBMIT) {
                    uri = submit_form(&awin->cd, in, &method);
                    if (!uri) return 0;
                }
            } else {
                /* User clicked on "nothing", so blur the focused control */
                SetFocus(awin->window->hwnd);
            }

            if (uri) {
                clientdata_destroy(&awin->cd);
                SetCursor(LoadCursor(NULL, IDC_APPSTARTING));
                anya_window_load(awin, uri, method);
                free(uri);
                return 0;
            }
        }
        break;
    case WM_CHAR:
        {
            /* Nothing is focused, so hand character to our "url bar" */
            static wchar_t buff[128];
            static int c = 0;
            wchar_t ch = (TCHAR)wParam;

            switch(ch) {
            case 13: /* return */
                if (c > 2) {
                    char uri[128];
                    wcstombs(uri, buff, sizeof(uri));
                    clientdata_destroy(&awin->cd);
                    anya_window_load(awin, uri, ANYA_REQUEST_GET);
                    c = 0;
                    buff[0] = L'\0';
                    return 0;
                }
                break;

            case 8: /* backspace */
                if (c > 0)
                    c--;
                buff[c] = L'\0';
                break;

            default: /* everything else */
                buff[c] = ch;
                buff[c + 1] = L'\0';
                c++;
                break;
            }
            oz_window_set_title(awin->window, buff);
            return 0;
        }
        break;

    case WM_PAINT:
        anya_window_draw(awin);
        return 0;

    case WM_HSCROLL:
        scroll_window(awin->window, wParam, SB_HORZ);
        return 0;
    case WM_VSCROLL:
        scroll_window(awin->window, wParam, SB_VERT);
        return 0;

    case WM_DESTROY:
#ifdef _WIN32_WCE
        CommandBar_Destroy(menuBar);
#endif
        PostQuitMessage(0);
        return 0;

#ifdef _WIN32_WCE
    case WM_CREATE:
        /* make menubar */
        {
            SHMENUBARINFO mbi;

            memset(&mbi, 0, sizeof(SHMENUBARINFO));
            mbi.cbSize     = sizeof(SHMENUBARINFO);
            mbi.hwndParent = hwnd;
            mbi.nToolBarId = 0;
            mbi.hInstRes   = GetModuleHandle(NULL);
            mbi.nBmpId     = 0;
            mbi.cBmpImages = 0;
            mbi.dwFlags = SHCMBF_EMPTYBAR;

            if (SHCreateMenuBar(&mbi)) 
                menuBar = mbi.hwndMB;
        }
        memset (&s_sai, 0, sizeof (s_sai));
        s_sai.cbSize = sizeof (s_sai);

        break;
    case WM_ACTIVATE:
        /* Notify shell of our activate message */
        SHHandleWMActivate(hwnd, wParam, lParam, &s_sai, FALSE);
        break;
    case WM_SETTINGCHANGE:
        SHHandleWMSettingChange(hwnd, wParam, lParam, &s_sai);
        break;
#endif
   }
   return DefWindowProcW(hwnd, message, wParam, lParam);
}
