/* -*- Mode: C++; tab-width: 20; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Anya client code
 *
 * The Initial Developer of the Original Code is Netscape Communications
 * Corporation. Portions created by the Initial Developer are
 * Copyright (C) 2003 the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *    Stuart Parmenter <pavlov@netscape.com>
 *    Joe Hewitt <hewitt@netscape.com>
 *    Ben Goodger <ben@netscape.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 */

/*
 * Client communications
 */

#include "anya.h"
#include "client.h"
#include "xcoder.h"
#include "prims.h"
#include "oz.h"
#include "idmap.h"

#if defined(_WIN32_WCE)
#include <winsock.h>
#elif defined(_WIN32)
#include <winsock2.h>
#elif defined(OZ_PALMOS)
extern UInt16 netLibRefNum; /* from main.c */
#endif

static int
receive_data(XCoderStream *stream, ClientData *cd)
{
    uint8 abc;
    uint8 *cp;
    uint8 *drawend;
    uint32 realsize;
    uint32 formpos = 0;

    /* read in the header */
    xcoder_Header(stream, &cd->header);

    if (cd->header.version == 0) {
        return 0;
    }

    /* allocate space for everything */
    if (cd->header.display_list_size > 0) {
        cd->drawing = malloc(cd->header.display_list_size);
        if (!cd->drawing)
            return 0;
    }

    if (cd->header.form_list_length > 0) {
        cd->forms = malloc(cd->header.form_list_length * sizeof(struct Input *));
        if (!cd->forms)
            return 0;
    }

    cp = cd->drawing;
    drawend = cp + cd->header.display_list_size;

    do {
        if (!xcoder_Read8(stream, &abc))
            abc = ABC_ERROR;

        switch(abc) {
        case ABC_IDMAP: /* this should be part of the header or
                           at least bail if we get multiple ones */
            xcoder_IdMap(stream, &cd->idmap);
            break;
        case ABC_POINT:
            xcoder_Point(stream, (struct Point *)cp);
            cp += sizeof(struct Point);
            break;
        case ABC_LINE:
            xcoder_Line(stream, (struct Line *)cp);
            cp += sizeof(struct Line);
            break;
        case ABC_RECT:
            xcoder_Rect(stream, (struct Rect *)cp);
            cp += sizeof(struct Rect);
            break;
        case ABC_ARC:
            xcoder_Arc(stream, (struct Arc *)cp);
            cp += sizeof(struct Arc);
            break;
        case ABC_POINTS:
            xcoder_Points(stream, (struct Points *)cp);
            cp += sizeof(struct Points);
            break;
        case ABC_POLYLINE:
            xcoder_Polyline(stream, (struct Polyline *)cp);
            cp += sizeof(struct Polyline);
            break;
        case ABC_POLYGON:
            xcoder_Polygon(stream, (struct Polygon *)cp);
            cp += sizeof(struct Polygon);
            break;
        case ABC_TEXT:
            xcoder_Text(stream, (struct Text *)cp);
            cp += sizeof(struct Text);
            break;
        case ABC_FONT:
            xcoder_Font(stream, (struct Font *)cp);
            cp += sizeof(struct Font);
            break;
        case ABC_COLOR:
            xcoder_Color(stream, (struct Color *)cp);
            cp += sizeof(struct Color);
            break;
        case ABC_IMAGE:
            xcoder_Image(stream, (struct Image *)cp);
            cp += sizeof(struct Image);
            break;

        case ABC_EVENT:
            drawend -= sizeof(struct Event);
            xcoder_Event(stream, (struct Event *)drawend);
            break;

        case ABC_HTMLELEMENT: /* (generic) html element */
            /*xcoder_HtmlElement(stream, (struct HtmlElement *)cp);*/
            cp += sizeof(struct HtmlElement);
            break;

        case ABC_INPUT_FORM:
            xcoder_InputForm(stream, (struct InputForm *)cp);
            cd->forms[formpos] = (struct Input *)cp;
            ++formpos;
            cp += sizeof(struct InputForm);
            break;
        case ABC_INPUT_HIDDEN:
            xcoder_InputHidden(stream, (struct InputHidden *)cp);
            cd->forms[formpos] = (struct Input *)cp;
            ++formpos;
            cp += sizeof(struct InputHidden);
            break;
        case ABC_INPUT_TEXT:
        case ABC_INPUT_PASSWORD:
        case ABC_INPUT_TEXTAREA:
            xcoder_InputText(stream, abc, (struct InputText *)cp);
            cd->forms[formpos] = (struct Input *)cp;
            ++formpos;
            cp += sizeof(struct InputText);
            break;
        case ABC_INPUT_CHECKBOX:
            xcoder_InputCheckbox(stream, (struct InputCheckbox *)cp);
            cd->forms[formpos] = (struct Input *)cp;
            ++formpos;
            cp += sizeof(struct InputCheckbox);
            break;
        case ABC_INPUT_RADIO:
            xcoder_InputRadio(stream, (struct InputRadio *)cp);
            cd->forms[formpos] = (struct Input *)cp;
            ++formpos;
            cp += sizeof(struct InputRadio);
            break;
        case ABC_INPUT_SUBMIT:
            xcoder_InputSubmit(stream, (struct InputSubmit *)cp);
            cd->forms[formpos] = (struct Input *)cp;
            ++formpos;
            cp += sizeof(struct InputSubmit);
            break;
        case ABC_INPUT_SELECT:
            xcoder_InputSelect(stream, (struct InputSelect *)cp);
            cd->forms[formpos] = (struct Input *)cp;
            ++formpos;
            cp += sizeof(struct InputSelect);
            break;

        default:
            break;
        }
    } while (abc != ABC_EOF && abc != ABC_ERROR);

    /* events are all put at the end of the drawing data, backwards,
       so that we can have fast access to them without having to walk
       through all the drawing data
     */
    cd->events = drawend;
    cd->event_list_size = (cd->drawing + cd->header.display_list_size) - drawend;

    /* compute the new display size (minus events, errors, etc) */
    realsize = cp - cd->drawing;

    if (cd->header.display_list_size != realsize + cd->event_list_size) {
        /* bail, don't leave anything here that can cause us to crash. */
        /* XXX this will cause us to leak anything that we were holding on to. */
        if (cd->drawing)
            free(cd->drawing);
        if (cd->forms)
            free(cd->forms);
        cd->drawing = NULL;
        cd->forms = NULL;
        cd->events = NULL;
        cd->event_list_size = 0;
        cd->header.form_list_length = 0;

        realsize = 0;
#if 0
        /* we probably hit an error and need to adjust the location
           of the events to be at the end of the drawing list so that
           we can easily walk it for freeing */
        memmove(cd->drawing + realsize, cd->events, cd->event_list_size);
        cd->events = cd->drawing + realsize;
#endif
    }
    cd->header.display_list_size = realsize;

    return (abc != ABC_ERROR);
}

static int
get_device_colors()
{
    switch(oz_screen_get_depth()) {
    case 8:
        return ANYA_DEVICE_COLOR256;
    case 15:
    case 16:
        return ANYA_DEVICE_COLOR16b;
    case 24:
        return ANYA_DEVICE_COLOR24b;
    case 32:
        return ANYA_DEVICE_COLOR32b;
    default:
        return 0;
    }
}

static int
send_request(const char *uri, int sock, uint8 method)
{
    /*
    purpose: initial connection, requests, form submission, dhtml, other stuff?

    header:
	    required:
		    Anya protocol version (8.8 fixed), dpi (8 bits), color/black&white/greyscale (8 bits),
            resolution (16bits x 16bits), unique device id? (32 bit)
	    optional:
		    send images on/off, other device specifics?

    request (think http get):
	    get URL/URI

    posting (think http post):
	    post url/uri + form arguments

    dhtml request:
	    figure out what stuff we need to do and build an extensible protocol for handling dynamic stuff


        -----------

    examples

    requesting Google:
    header:	 0.01, 92, color, 320, 180, 000000001
    request: GET "http://www.google.com/" "\0"

    */
    int ret;
    XCoderStream writer;

    struct Request r;
    r.version = ANYA_VERSION;
    r.request = method;
    r.device.display = get_device_colors();
    r.device.dpi = oz_screen_get_dpi();
    r.device.xres = oz_screen_get_width();
    r.device.yres = oz_screen_get_height();
    r.device.uid = 1;
    r.uri = (char *)uri; /* we promise not to write to this */

    /* open xcoder for writing */
    xcoder_Open(&writer, sock, XCODER_WRITE);

    ret = (xcoder_Request(&writer, &r) &&
           xcoder_FlushBuffer(&writer));

    xcoder_Close(&writer);

    return ret;
}

static int
get_server_socket()
{
#ifdef OZ_PALMOS
    NetSocketRef sock;
    NetHostInfoType *hp;
    NetHostInfoBufType hpbuf;
    NetSocketAddrINType addr;
    Err err;

    sock = NetLibSocketOpen(netLibRefNum,
                            netSocketAddrINET, netSocketTypeStream, netSocketProtoIPTCP,
                            -1, &err);
    if (sock < 0)
       return -1;

    hp = NetLibGetHostByName(netLibRefNum,
                             ANYA_SERVER, (NetHostInfoBufPtr)&hpbuf,
                             -1, &err);
    if (!hp)
      return -1;

    MemSet(&addr, sizeof(addr), 0);
    addr.family = netSocketAddrINET;
    addr.port = NetHToNS(ANYA_PORT);
    MemMove(&addr.addr, hp->addrListP[0], hp->addrLen);

    if (NetLibSocketConnect(netLibRefNum,
                            sock, (NetSocketAddrType*)&addr, sizeof(addr),
                            -1, &err) >= 0)
        return sock;
        
    return -1;
#else
    int sock;
    struct hostent* hp;
    struct sockaddr_in addr;

    sock = socket(AF_INET, SOCK_STREAM, IPPROTO_IP);
    if (sock < 0)
        return -1;

    hp = gethostbyname(ANYA_SERVER);
    if (!hp)
        return -1;

    memset(&addr, 0, sizeof(addr));
    addr.sin_family = AF_INET;
    addr.sin_port = htons(ANYA_PORT);
    memcpy(&addr.sin_addr, hp->h_addr, hp->h_length);

    if (connect(sock, (struct sockaddr*)&addr, sizeof(addr)) >= 0)
        return sock;
    
    return -1;
#endif
}

int
clientdata_request(ClientData *cd, const char *uri, uint8 method)
{
    int sock;
    int ret;

    XCoderStream reader;
    
    sock = get_server_socket();
    if (sock >= 0 &&
        send_request(uri, sock, method)) {
        /* open xcoder for reading */
        xcoder_Open(&reader, sock, XCODER_READ /* | XCODER_COMPRESS */);
        ret = receive_data(&reader, cd);
        xcoder_Close(&reader);
    }
#ifdef OZ_PALMOS
    NetLibSocketClose(0, sock, -1, NULL);
#else
    closesocket(sock);
#endif
    return ret;

}

static void
free_HtmlElement(struct HtmlElement *element)
{
    uint32 i;
    for (i = 0; i < element->nattrs; ++i) {
        if ((element->attrs[i].value & 1) &&
            ((element->attrs[i].value >> 1) > 0))
                free(element->attrs[i].values);

    }
    for (i = 0; i < element->nkids; ++i)
        free_HtmlElement(&element->kids[i]);

    if (element->nkids)
        free(element->kids);
}

void
clientdata_destroy(ClientData *cd)
{
    uint8 *cp = cd->drawing;
    const uint8 *end = cp + cd->header.display_list_size + cd->event_list_size;

    if (cd->header.version == 0)
        return;

    while (cp < end) {

        switch (*cp) {

        case ABC_POINT:
            cp += sizeof(struct Point);
            break;
        case ABC_LINE:
            cp += sizeof(struct Line);
            break;
        case ABC_RECT:
            cp += sizeof(struct Rect);
            break;
        case ABC_ARC:
            cp += sizeof(struct Arc);
            break;
        case ABC_POINTS:
            {
                struct Points *points = (struct Points *)cp;
                if (points->npoints > 0)
                    free(points->points);
                cp += sizeof(struct Polygon);
            }
            break;
        case ABC_POLYLINE:
            {
                struct Polyline *poly = (struct Polyline *)cp;
                if (poly->npoints > 0)
                    free(poly->points);
                cp += sizeof(struct Polyline);
            }
            break;
        case ABC_POLYGON:
            {
                struct Polygon *poly = (struct Polygon *)cp;
                if (poly->npoints > 0)
                    free(poly->points);
                cp += sizeof(struct Polygon);
            }
            break;
        case ABC_TEXT:
            cp += sizeof(struct Text);
            break;
        case ABC_FONT:
            cp += sizeof(struct Font);
            break;
        case ABC_COLOR:
            cp += sizeof(struct Color);
            break;

        case ABC_IMAGE:
            {
                struct Image *image = (struct Image *)cp;
                if (image->nbytes > 0)
                    free(image->bytes);
                cp += sizeof(struct Image);
            }
            break;

        case ABC_EVENT:
            {
                struct Event *event = (struct Event *)cp;
                if (event->nrects > 0)
                    free(event->rects);
                cp += sizeof(struct Event);
            }
            break;

        case ABC_HTMLELEMENT:
            /* free_HtmlElement((struct HtmlElement *)cp); */
            cp += sizeof(struct HtmlElement);
            break;

        case ABC_INPUT_FORM:
            cp += sizeof(struct InputForm);
            break;
        case ABC_INPUT_HIDDEN:
            cp += sizeof(struct InputHidden);
            break;
        case ABC_INPUT_TEXT:
        case ABC_INPUT_TEXTAREA:
        case ABC_INPUT_PASSWORD:
            {
                struct InputText *in = (struct InputText *)cp;
#ifdef _WIN32
                if (in->data)
                    DestroyWindow(in->data);
#endif
                cp += sizeof(struct InputText);
            }
            break;
        case ABC_INPUT_CHECKBOX:
            {
                struct InputCheckbox *in = (struct InputCheckbox *)cp;
#ifdef _WIN32
                if (in->data)
                    DestroyWindow(in->data);
#endif
                cp += sizeof(struct InputCheckbox);
            }
            break;
        case ABC_INPUT_RADIO:
            {
                struct InputRadio *in = (struct InputRadio *)cp;
#ifdef _WIN32
                if (in->data)
                    DestroyWindow(in->data);
#endif
                cp += sizeof(struct InputRadio);
            }
            break;
        case ABC_INPUT_SUBMIT:
            cp += sizeof(struct InputSubmit);
            break;
        case ABC_INPUT_SELECT:
            {
                struct InputSelect *in = (struct InputSelect *)cp;
                if (in->options)
                    free(in->options);
#ifdef _WIN32
                if (in->data)
                    DestroyWindow(in->data);
#endif
                cp += sizeof(struct InputSelect);
            }
            break;

        default:
            break;
        }
    }

    /* XXX need to free html elements */

    idmap_Destroy(&cd->idmap);

    if (cd->header.display_list_size > 0)
        free(cd->drawing);

    if (cd->header.form_list_length > 0)
        free(cd->forms);
}
