/* -*- Mode: C++; tab-width: 20; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Anya client code
 *
 * The Initial Developer of the Original Code is Netscape Communications
 * Corporation. Portions created by the Initial Developer are
 * Copyright (C) 2003 the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *    Stuart Parmenter <pavlov@netscape.com>
 *    Joe Hewitt <hewitt@netscape.com>
 *    Ben Goodger <ben@netscape.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 */

/*
 * Common controls
 */

#include "controls.h"
#include "oz.h"
#include "utf8.h"

#include <commctrl.h>

#if defined(_WIN32_WCE)
#define assert(x)
#elif defined(_WIN32)
#include <assert.h>
#endif


#define INPUT_PADDING 2
#define MAX_INPUT_TEXT 1000

static void
create_text_widget(AnyaWindow *awin, struct InputText *text)
{
    if (!text->data) {
        HFONT font;
        uint32 stringlen;
        int type = 0;

        const char *string = idmap_GetString(&awin->cd.idmap, text->value, &stringlen);
        int wlen = sizeof(wchar_t) * (stringlen + 1);
        wchar_t *wtext = malloc(wlen);
        utf8_to_ucs2(string, &stringlen, wtext, &wlen);
        wtext[wlen] = L'\0';

        if (text->i.type == ABC_INPUT_TEXTAREA)
            type = ES_MULTILINE;
        else if (text->i.type == ABC_INPUT_PASSWORD)
            type = ES_PASSWORD;

        text->data = CreateWindowExW(0L, L"EDIT", wtext,
                              WS_VISIBLE | WS_CHILD | type,
                              text->x+INPUT_PADDING, text->y+INPUT_PADDING,
                              text->w-INPUT_PADDING*2, text->h-INPUT_PADDING*2,
                              awin->window->hwnd, 
                              (HMENU) text, NULL, NULL);
        free(wtext);

        if (text->maxlength > 0)
            SendMessageW(text->data, EM_SETLIMITTEXT, (WPARAM) text->maxlength, (LPARAM) NULL);

#ifdef _WIN32_WCE
        font = (HFONT)GetStockObject(SYSTEM_FONT);
#else
        font = (HFONT)GetStockObject(ANSI_VAR_FONT);
#endif
        SendMessageW(text->data, WM_SETFONT, (WPARAM) font, (LPARAM) NULL);
        DeleteObject(font);

    }
}

static void
create_checkbox_widget(AnyaWindow *awin, struct InputCheckbox *cbx)
{
    if (!cbx->data) {
        cbx->data = CreateWindowExW(0L, L"BUTTON", NULL,
                        WS_VISIBLE | WS_CHILD | BS_AUTOCHECKBOX,
                        cbx->x, cbx->y, cbx->w, cbx->h,
                        awin->window->hwnd, 
                         (HMENU) cbx, GetModuleHandle(NULL), NULL);

        /* Set the checked value of the control */
        if (cbx->checked)
            SendMessageW(cbx->data, BM_SETCHECK, (WPARAM) BST_CHECKED, (LPARAM) 0);
    }
}

static void
create_radio_widget(AnyaWindow *awin, struct InputRadio *radio)
{
    if (!radio->data) {
        radio->data = CreateWindowExW(0L, L"BUTTON", NULL,
                        WS_VISIBLE | WS_CHILD | BS_NOTIFY | BS_RADIOBUTTON,
                        radio->x, radio->y, radio->w, radio->h,
                        awin->window->hwnd, 
                        (HMENU) radio, GetModuleHandle(NULL), NULL);

        oz_set_hwnd_userdata(radio->data, radio);

        /* Set the checked value of the control */
        if (radio->checked)
            SendMessageW(radio->data, BM_SETCHECK, (WPARAM) BST_CHECKED, (LPARAM) 0);

    }
}

static void
create_select_widget(AnyaWindow *awin, struct InputSelect *sel)
{
    if (!sel->data) {
        uint32 i;
        HFONT font;

        if (sel->iscombo) {
            sel->data = CreateWindowExW(0L, L"COMBOBOX", NULL,
                                  WS_VISIBLE | WS_CHILD | LBS_NOTIFY | CBS_DROPDOWNLIST,
                                  sel->x, sel->y, sel->w, sel->h+150,
                                  awin->window->hwnd, 
                                  0, GetModuleHandle(NULL), NULL);
        } else {
            sel->data = CreateWindowExW(0L, L"LISTBOX", NULL,
                                  WS_VISIBLE | WS_CHILD | LBS_NOTIFY | WS_VSCROLL,
                                  sel->x+INPUT_PADDING, sel->y+INPUT_PADDING,
                                  sel->w-INPUT_PADDING*2, sel->h-INPUT_PADDING*2,
                                  awin->window->hwnd, 
                                  0, GetModuleHandle(NULL), NULL);
        }

#ifdef _WIN32_WCE
        font = (HFONT)GetStockObject(SYSTEM_FONT);
#else
        font = (HFONT)GetStockObject(ANSI_VAR_FONT);
#endif
        SendMessageW(sel->data, WM_SETFONT, (WPARAM) font, (LPARAM) NULL);
        DeleteObject(font);

        /* Populate with options */
        for (i = 0; i < sel->noptions; ++i) {
            uint32 stringlen;
            const char *string = idmap_GetString(&awin->cd.idmap, sel->options[i].label, &stringlen);
            if (string) {
                int wlen = sizeof(wchar_t) * (stringlen + 1);
                wchar_t *wtext = malloc(wlen);
                utf8_to_ucs2(string, &stringlen, wtext, &wlen);
                wtext[wlen] = L'\0';
                SendMessageW(sel->data, sel->iscombo ? CB_ADDSTRING : LB_ADDSTRING, 0, (LPARAM)wtext); 
                free(wtext);
            }

            if (i == sel->selectedindex) {
                if (sel->iscombo)
                    SendMessageW(sel->data, CB_SETCURSEL, i, 0);
                else
                    SendMessageW(sel->data, LB_SETCURSEL, i, 0);
            }
        }
    }
}

void
anya_init_controls()
{
    /* Init win32 common controls */
    INITCOMMONCONTROLSEX icex;
    icex.dwSize = sizeof(INITCOMMONCONTROLSEX);
    icex.dwICC  = ICC_LISTVIEW_CLASSES;
    InitCommonControlsEx(&icex); 
}

void
anya_create_controls(AnyaWindow *awin)
{
    uint32 i;
    for (i = 0; i < awin->cd.header.form_list_length; ++i) {
        struct Input *input = awin->cd.forms[i];
        switch (input->type) {
        case ABC_INPUT_TEXT:
        case ABC_INPUT_PASSWORD:
        case ABC_INPUT_TEXTAREA:
            {
                struct InputText *text = (struct InputText *)input;
                create_text_widget(awin, text);
            }
            break;
        case ABC_INPUT_SELECT:
            {
                struct InputSelect *sel = (struct InputSelect *)input;
                create_select_widget(awin, sel);
            }
            break;
        case ABC_INPUT_CHECKBOX:
            {
                struct InputCheckbox *cbx = (struct InputCheckbox *)input;
                create_checkbox_widget(awin, cbx);
            }
            break;
        case ABC_INPUT_RADIO:
            {
                struct InputRadio *radio = (struct InputRadio *)input;
                create_radio_widget(awin, radio);
            }
            break;
        default:
            break;
        }
    }
}

void
anya_handle_radio_click(AnyaWindow *awin, struct InputRadio *radio)
{
    uint32 i;
    uint8 checked;
    uint32 radioNameLen;
    const char *radioName = 0;

    /* Look for a checked radio button in the same group as the clicked one */
    for (i = 0; i < awin->cd.header.form_list_length; ++i) {
        struct Input *elem = awin->cd.forms[i];
        switch(elem->type) {
        case ABC_INPUT_RADIO:
            {
                struct InputRadio *radio2 = (struct InputRadio *)elem;
                if (radio2->checked) {
                    /* Compare the group names of the two radio buttons */
                    const char *radio2Name = idmap_GetString(&awin->cd.idmap, radio2->name, &radioNameLen);
                    if (!radioName) 
                        radioName = idmap_GetString(&awin->cd.idmap, radio->name, &radioNameLen);

                    if (strcmp(radioName, radio2Name) == 0) {
                        /* The radios are in the same group, so uncheck the checked one */
                        SendMessageW(radio2->data, BM_SETCHECK, (WPARAM)0, (LPARAM) 0);
                        radio2->checked = 0;
                    }
                }
            }
            break;
        default:
            break;
        }
    }

    /* Check the clicked radio button */
    checked = (uint8)SendMessageW(radio->data, BM_GETCHECK, (WPARAM)0, (LPARAM) 0);
    if (!checked) {
        SendMessageW(radio->data, BM_SETCHECK, (WPARAM)1, (LPARAM) 0);
        radio->checked = 1;
    }
}

static int
append_name_and_value(char *buffer, int pos,
                      const char *name, uint32 namelen,
                      const char *value, uint32 valuelen)
{
    int initalpos = pos;

    if (!name)
        return 0;

    strncpy(buffer + pos, name, namelen);
    pos += namelen;
    *(buffer + pos) = '=';
    pos += 1;
    strncpy(buffer + pos, value, valuelen);
    pos += valuelen;
    *(buffer + pos) = '&';
    pos += 1;

    return pos - initalpos;
}
                      

/* XXX this function should be renamed and only used to build a
       URL to submit via one of the _load functions */
char *
submit_form(const ClientData *cd, const struct Input *submit, uint8 *method)
{
    char *ret;
    struct Input *elem;
    uint32 i;
    char url[1024]; /* XXX urlpos becomes > 1024, we need to malloc a larger buffer */
    int urlpos = 0;

    if (!cd->forms)
        return NULL;

    for (i = 0; i < cd->header.form_list_length; ++i) {
        elem = cd->forms[i];

        if (submit->formid != elem->formid)
            continue;

        switch(elem->type) {
        case ABC_INPUT_FORM:
            {
                const struct InputForm *form = (const struct InputForm *)elem;
                const char *action;
                uint32 actionlen;

                assert(urlpos == 0);

                /* set the method */
                *method = (form->method == 0) ? ANYA_REQUEST_GET : ANYA_REQUEST_POST;

                action = idmap_GetString(&cd->idmap, form->action, &actionlen);

                strncpy(url + urlpos, action, actionlen);
                urlpos += actionlen;
                strncpy(url + urlpos, "?", 1);
                urlpos += 1;
            }
            break;
        case ABC_INPUT_HIDDEN:
            {
                const struct InputHidden *hid = (const struct InputHidden *)elem;
                uint32 namelen;
                uint32 valuelen;
                const char *name = idmap_GetString(&cd->idmap, hid->name, &namelen);
                const char *value = idmap_GetString(&cd->idmap, hid->value, &valuelen);

                urlpos += append_name_and_value(url, urlpos, name, namelen, value, valuelen);
            }
            break;
        case ABC_INPUT_TEXT:
        case ABC_INPUT_PASSWORD:
        case ABC_INPUT_TEXTAREA:
            {
                const struct InputText *text = (const struct InputText *)elem;
                uint32 namelen;
                uint32 valuelen;
                const char *name = idmap_GetString(&cd->idmap, text->name, &namelen);
                const char *value = NULL;
                char buf[256];

                if (text->data) {
                    /* We have to read the text from the control line by line */
                    uint32 i;
                    uint32 linecount = SendMessageW(text->data, EM_GETLINECOUNT, 0, 0);
                    valuelen = SendMessageW(text->data, WM_GETTEXTLENGTH, 0, 0);

                    if (valuelen > 0) {
                        wchar_t wbuf[256];

                        /* XXX if valuelen < sizeof(buffer) we should malloc a bigger buffer.
                           we use the stack buffer by default since it is more efficient */
                        assert(valuelen < sizeof(wbuf) / sizeof(wchar_t));

                        for (i = 0; i < linecount; ++i) {
                            uint32 linelen = SendMessageW(text->data, EM_LINELENGTH, 0, 0);
                            /* first byte needs to contain the size of our buffer in
                               characters */
                            ((WORD*)wbuf)[0] = sizeof(wbuf) / sizeof(wchar_t);
                            SendMessageW(text->data, EM_GETLINE, (WPARAM) 0, (LPARAM)wbuf);
                        }

                        wbuf[valuelen] = L'\0';
                    
                        wcstombs(buf, wbuf, sizeof(buf));

                        value = buf;
                    }
                } else {
                    value = idmap_GetString(&cd->idmap, text->value, &valuelen);
                }

                urlpos += append_name_and_value(url, urlpos, name, namelen, value, valuelen);
            }
            break;
         case ABC_INPUT_CHECKBOX:
            {
                const struct InputCheckbox *cbx = (const struct InputCheckbox *)elem;
                uint32 namelen;
                const char *name = idmap_GetString(&cd->idmap, cbx->name, &namelen);
                uint8 checked;

                if (cbx->data) {
                    checked = (uint8)SendMessageW(cbx->data, BM_GETCHECK, 0, 0);
                } else {
                    checked = cbx->checked;
                }

                if (checked)
                    urlpos += append_name_and_value(url, urlpos, name, namelen, "on", 2);
            }
            break;
         case ABC_INPUT_RADIO:
            {
                const struct InputRadio *radio = (const struct InputRadio *)elem;
                uint32 namelen;
                const char *name = idmap_GetString(&cd->idmap, radio->name, &namelen);
                uint8 checked;

                if (radio->data) {
                    checked = (uint8)SendMessageW(radio->data, BM_GETCHECK, 0, 0);
                } else {
                    checked = radio->checked;
                }

                if (checked) {
                    uint32 valuelen;
                    const char *value = idmap_GetString(&cd->idmap, radio->value, &valuelen);
                    urlpos += append_name_and_value(url, urlpos, name, namelen, value, valuelen);
                }
            }
            break;
        case ABC_INPUT_SELECT:
            {
                const struct InputSelect *select = (const struct InputSelect *)elem;
                const char *name;
                uint32 namelen;
                const char *value;
                uint32 valuelen;
                uint32 selindex;

                if (select->iscombo) {
                    selindex = SendMessageW(select->data, CB_GETCURSEL, 0, 0);
                } else {
                    selindex = SendMessageW(select->data, LB_GETCURSEL, 0, 0);
                }
 
                name = idmap_GetString(&cd->idmap, select->name, &namelen);
                value = idmap_GetString(&cd->idmap, select->options[selindex].value, &valuelen);
                urlpos += append_name_and_value(url, urlpos, name, namelen, value, valuelen);
            }
            break;
        case ABC_INPUT_SUBMIT:
            {
                const struct InputSubmit *sub = (const struct InputSubmit *)elem;
                uint32 namelen;
                uint32 valuelen;
                const char *name;
                const char *value;

                /* if we're a submit button that isn't the one that was clicked on,
                   don't append anything */
                if ((const struct Input *)sub != submit)
                    break;

                name = idmap_GetString(&cd->idmap, sub->name, &namelen);
                value = idmap_GetString(&cd->idmap, sub->value, &valuelen);

                urlpos += append_name_and_value(url, urlpos, name, namelen, value, valuelen);
            }
            break;
        default:
            /* assert */
            break;
        }
    }

    if (url[urlpos - 1] == '&')
        urlpos--;

    ret = malloc(urlpos + 1);
    if (!ret)
        return NULL;
    strncpy(ret, url, urlpos);
    ret[urlpos] = '\0';

    return ret;
}
