/* -*- Mode: C++; tab-width: 20; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Anya client code
 *
 * The Initial Developer of the Original Code is Netscape Communications
 * Corporation. Portions created by the Initial Developer are
 * Copyright (C) 2002-2003 the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *    Stuart Parmenter <pavlov@netscape.com>
 *    Ben Goodger <ben@netscape.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 */

/*
 * Image drawing code
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "oz.h"
#include "ozprivate.h"

#ifndef _WIN32_WCE
#include <assert.h>
#else
#define assert(x)
#endif

typedef struct {
    BITMAPINFOHEADER bmiHeader;
    union {
        RGBQUAD      bmiColors[256];
        DWORD        bmiFields[256];
    };
} ozBITMAPINFO;

#define WEBSAFE_PALETTE_SIZE 216 * sizeof(RGBQUAD)
/* XXX maybe convert this to a static const array to save malloc'ing
       also, need to free this at shutdown */
static RGBQUAD *gWebSafePalette = NULL;

static void
make_websafe_palette()
{
    int r, g, b;
    BYTE *cp;

    gWebSafePalette = malloc(WEBSAFE_PALETTE_SIZE);

    cp = (BYTE*)gWebSafePalette;

    for (r = 0; r < 6; ++r) {
        for (g = 0; g < 6; ++g) {
            for (b = 0; b < 6; ++b) {
                *cp++ = (r * 65535 / 5);
                *cp++ = (g * 65535 / 5);
                *cp++ = (b * 65535 / 5);
                *cp++ = 0;
            }
        }
    }
}

void
oz_draw_image(const ozWindow *window, const ozGC *gc,
              int x, int y, int width, int height,
              int destwidth, int destheight,
              int xoffset, int yoffset,
              uint8 istile, uint8 bpp, uint32 nbytes, const uint8 *bytes)
{
    uint32 bpl, imgSize;
    ozBITMAPINFO bmi;
    DWORD rop = SRCCOPY;
    HDC imageDC;
    HBITMAP imageBitmap, oldBitmap;
    uint8 *dibbits = NULL;

    bmi.bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
    bmi.bmiHeader.biWidth = width;
    bmi.bmiHeader.biHeight = height;
    bmi.bmiHeader.biPlanes = 1;
    bmi.bmiHeader.biCompression = BI_RGB;
    bmi.bmiHeader.biXPelsPerMeter = 0;
    bmi.bmiHeader.biYPelsPerMeter = 0;
    bmi.bmiHeader.biClrUsed = 0;
    bmi.bmiHeader.biClrImportant = 0;

    bpl = (width * bpp * 8) >> 5;
    if ((width * bpp * 8) & 0x1f)
        ++bpl;
    bpl <<= 2;

    imgSize = bpl * height;

    imageDC = CreateCompatibleDC(gc->dc);

    if (imgSize < nbytes) {
        /* We have alpha data too, extract it and use it as a mask */
        const uint8 *alphaBytes = bytes + imgSize;
        const size_t alphaSize = nbytes - imgSize;
        HBITMAP alphaBitmap;

        bmi.bmiHeader.biBitCount = 1;
        bmi.bmiHeader.biSizeImage = alphaSize;

        /* Note that the palette is being set up so the DIB and the DDB have white and
           black reversed. This is because we need the mask to have 0 for the opaque
           pixels of the image, and 1 for the transparent pixels. This way the SRCAND
           operation sets the opaque pixels to 0, and leaves the transparent pixels
           undisturbed */
        bmi.bmiColors[0].rgbBlue = 255;
        bmi.bmiColors[0].rgbGreen = 255;
        bmi.bmiColors[0].rgbRed = 255;
        bmi.bmiColors[0].rgbReserved = 0;
        bmi.bmiColors[1].rgbBlue = 0;
        bmi.bmiColors[1].rgbGreen = 0;
        bmi.bmiColors[1].rgbRed = 0;
        bmi.bmiColors[1].rgbReserved = 0;

        alphaBitmap = CreateDIBSection(gc->dc, (LPBITMAPINFO)&bmi, DIB_RGB_COLORS, &dibbits, NULL, 0);
        assert(alphaBitmap);

        if (alphaBitmap) {
            memcpy(dibbits, alphaBytes, nbytes - imgSize);

            oldBitmap = SelectObject(imageDC, alphaBitmap);

            StretchBlt(gc->dc, x, y, destwidth, destheight, imageDC, 0, 0, width, height, SRCAND);

            SelectObject(imageDC, oldBitmap);
            DeleteObject(alphaBitmap);
        }

        rop = SRCPAINT;
    }

    bmi.bmiHeader.biBitCount = bpp * 8;
    bmi.bmiHeader.biSizeImage = imgSize;

    if (bmi.bmiHeader.biBitCount == 8) {
        if (!gWebSafePalette)
            make_websafe_palette();

        memcpy(bmi.bmiColors, gWebSafePalette, WEBSAFE_PALETTE_SIZE);
    } else if (bmi.bmiHeader.biBitCount == 16) {
        /* Specify masks for 555 image data */
        bmi.bmiHeader.biBitCount = 16;
        bmi.bmiHeader.biCompression = BI_BITFIELDS;

        bmi.bmiFields[0] = 0x7C00;
        bmi.bmiFields[1] = 0x03E0;
        bmi.bmiFields[2] = 0x001F;
    }

    imageBitmap = CreateDIBSection(gc->dc, (LPBITMAPINFO)&bmi, DIB_RGB_COLORS, &dibbits, NULL, 0);
    assert(imageBitmap);

    if (imageBitmap) {
        memcpy(dibbits, bytes, imgSize);
        oldBitmap = SelectObject(imageDC, imageBitmap);

        StretchBlt(gc->dc, x, y, destwidth, destheight, imageDC, 0, 0, width, height, rop);

        SelectObject(imageDC, oldBitmap);
        DeleteObject(imageBitmap);
    }

    DeleteDC(imageDC);
}

