/* -*- Mode: C++; tab-width: 20; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Anya client code
 *
 * The Initial Developer of the Original Code is Netscape Communications
 * Corporation. Portions created by the Initial Developer are
 * Copyright (C) 2002-2003 the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *    Stuart Parmenter <pavlov@netscape.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 */

/*
 * Windowing code
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "oz.h"
#include "ozprivate.h"

#include "utf8.h"

#ifdef _WIN32_WCE
#define OZ_WINDOW_STYLE    WS_HSCROLL | WS_VSCROLL
#define OZ_WINDOW_EX_STYLE 0
#else
#define OZ_WINDOW_STYLE    WS_OVERLAPPEDWINDOW | WS_HSCROLL | WS_VSCROLL
#define OZ_WINDOW_EX_STYLE WS_EX_APPWINDOW
#endif

ozWindow *
oz_window_create(ozWindow *parent, int x, int y, int width, int height)
{
    ozWindow *w = malloc(sizeof(ozWindow));
    if (!w)
        return NULL;

    w->parent = parent;

    w->x = x;
    w->y = y;
    w->width = width;
    w->height = height;

    w->hidden = TRUE;

    w->owner = NULL;

    w->hwnd = CreateWindowExW(OZ_WINDOW_EX_STYLE,
                              L"AnyaWndClass",
                              L"Main Window",
                              OZ_WINDOW_STYLE,
                              x, y,
                              width, height,
                              parent ? parent->hwnd : NULL,
                              (HMENU) NULL,
                              GetModuleHandle(NULL),
                              NULL);

    /* use userdata to store a pointer to w */
    oz_set_hwnd_userdata(w->hwnd, w);

    return w;
}

void
oz_window_destroy(ozWindow *win)
{
    oz_set_hwnd_userdata(win->hwnd, NULL);
    DestroyWindow(win->hwnd);
    free(win);
}

ozWindow *
oz_window_get_parent(ozWindow *win)
{
    return win->parent;
}

void
oz_window_move(ozWindow *win, int x, int y)
{
    win->x = x;
    win->y = y;
}

void
oz_window_resize(ozWindow *win, int width, int height)
{
    win->width = width;
    win->height = height;
}

void
oz_window_show(ozWindow *win)
{
    win->hidden = FALSE;
    ShowWindow(win->hwnd, SW_SHOWNORMAL);
    UpdateWindow(win->hwnd);
}

void
oz_window_hide(ozWindow *win)
{
    win->hidden = TRUE;
    ShowWindow(win->hwnd, SW_HIDE);
}

void
oz_window_set_title(ozWindow *win, const wchar_t *title)
{
    SendMessageW(win->hwnd, WM_SETTEXT, (WPARAM)0, (LPARAM)title);
}

void
oz_window_redraw(ozWindow *win)
{
    InvalidateRect(win->hwnd, NULL, TRUE);
    UpdateWindow(win->hwnd);
}

void
oz_window_scroll_set_range(ozWindow *win, int xmin, int xmax, int ymin, int ymax)
{
    SCROLLINFO si;
    RECT clientRect;

    GetClientRect(win->hwnd, &clientRect);

    si.cbSize = sizeof(si);
    si.fMask  = SIF_DISABLENOSCROLL | SIF_RANGE | SIF_PAGE;
    si.nPage  = 1;
    si.nMin   = xmin;
    si.nMax   = xmax - (clientRect.right-clientRect.left);
    SetScrollInfo(win->hwnd, SB_HORZ, &si, TRUE); 
    si.nMin   = ymin;
    si.nMax   = ymax - (clientRect.bottom-clientRect.top);
    
    SetScrollInfo(win->hwnd, SB_VERT, &si, TRUE);
}

void
oz_window_scroll_set_pos(ozWindow *win, int x, int y)
{
    SCROLLINFO si;
    si.cbSize = sizeof(SCROLLINFO);
    si.fMask = SIF_POS;
    si.nPos = x;
    SetScrollInfo(win->hwnd, SB_HORZ, &si, TRUE);
    si.nPos = y;
    SetScrollInfo(win->hwnd, SB_VERT, &si, TRUE);
}

void
oz_window_scroll_get_pos(const ozWindow *win, uint16 *x, uint16 *y)
{
    SCROLLINFO si;
    si.cbSize = sizeof(SCROLLINFO);
    si.fMask = SIF_POS;
    GetScrollInfo(win->hwnd, SB_HORZ, &si);
    *x = si.nPos;
    GetScrollInfo(win->hwnd, SB_VERT, &si);
    *y = si.nPos;
}

void
oz_window_set_owner(ozWindow *win, void *owner)
{
    win->owner = owner;
}

void *
oz_window_get_owner(ozWindow *win)
{
    if (!win)
        return NULL;
    return win->owner;
}

ozWindow *
oz_window_from_hwnd(HWND hwnd)
{
    return (ozWindow *)oz_get_hwnd_userdata(hwnd);
}

