/* -*- Mode: C++; tab-width: 20; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Anya server code
 *
 * The Initial Developer of the Original Code is Netscape Communications
 * Corporation. Portions created by the Initial Developer are
 * Copyright (C) 2003 the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *    Joe Hewitt <hewitt@netscape.com>
 *    Stuart Parmenter <pavlov@netscape.com>
 *    Ben Goodger <ben@netscape.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 */

#include "anyaBrowser.h"
#include "anyaDirProvider.h"

#include "xcoder.h"
#include "prims.h"
#include "abc.h"
#include "anyaIDataTransferManager.h"

#include "nsIGenericFactory.h"
#include "nsIComponentManager.h"
#include "nsCWebBrowser.h"
#include "nsIWebBrowserSetup.h"
#include "nsIWebNavigation.h"
#include "nsEmbedAPI.h"
#include "nsIProfile.h"
#include "nsString.h"
#include "nsIWidget.h"
#include "nsIDocShell.h"
#include "nsIPresShell.h"
#include "nsIPresContext.h"
#include "nsIViewManager.h"
#include "nsIView.h"
#include "nsIDOMDocument.h"
#include "nsIDocument.h"
#include "nsIDOMNodeList.h"
#include "nsIMIMEInputStream.h"
#include "nsIContent.h"
#include "nsIFrame.h"
#include "nsILink.h"
#include "nsIImageFrame.h"
#include "imgIRequest.h"
#include "imgIContainer.h"
#include "nsWidgetsCID.h"
#include "nsLocalFile.h"
#include "nsIHTMLContent.h"
#include "nsIDOMElement.h"
#include "nsIDOMHTMLFormElement.h"
#include "nsIDOMHTMLMapElement.h"
#include "nsIDOMHTMLCollection.h"
#include "nsIDOMHTMLInputElement.h"
#include "nsIDOMHTMLSelectElement.h"
#include "nsIDOMHTMLTextAreaElement.h"
#include "nsIDOMHTMLOptionElement.h"
#include "nsIForm.h"
#include "nsIFormControl.h"
#include "nsHTMLValue.h"
#include "nsImageMapUtils.h"
#include "gfxIImageFrame.h"
#include "nsIImage.h"
#include "nsReadableUtils.h"
#include "nsIURI.h"
#include "nsNetUtil.h"

#include <io.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

static NS_DEFINE_CID(kAppShellCID, NS_APPSHELL_CID);

//////////////////////////////////////////////

static anyaBrowser *gBrowser;

static nsresult SelectDefaultProfile();
static nsresult GetViewportView(nsIWebBrowser *aWebBrowser, nsIView **aView);
static nsresult GetScrollPortView(nsIWebBrowser *aWebBrowser, nsIView **aView);
static nsresult GetCanvasFrame(nsIWebBrowser *aWebBrowser, nsIFrame **aFrame);
static nsresult DumpForms(nsIDOMDocument *aDomDoc, nsIDocShell* aDocShell,
                          XCoderStream *aStream, unsigned long *aSize);
static nsresult DumpAnchors(nsIDOMDocument *aDomDoc, nsIDocShell* aDocShell,
                            XCoderStream *aStream, unsigned long *aSize);
static nsresult DumpImageMaps(nsIDOMDocument* aDomDoc, nsIDocShell* aDocShell, 
                              XCoderStream* aStream, unsigned long* aSize);
static nsresult GetImageRect(nsIDOMElement* aImageElement, nsIDocShell* aDocShell, 
                             ozRect* aImageRect);
static void FindBoundingBoxForRectArea(const nsAString& aCoords, ozRect* aBoundingBox);
static void FindBoundingBoxForPolyArea(const nsAString& aCoords, ozRect* aBoundingBox);
static void FindBoundingBoxForCircleArea(const nsAString& aCoords, ozRect* aBoundingBox);
static nscoord* lo_parse_coord_list(char *str, PRInt32* value_cnt);


//////////////////////////////////////////////

NS_IMPL_ADDREF(anyaBrowser)
NS_IMPL_RELEASE(anyaBrowser)

NS_INTERFACE_MAP_BEGIN(anyaBrowser)
   NS_INTERFACE_MAP_ENTRY_AMBIGUOUS(nsISupports, nsIWebBrowserChrome)
   NS_INTERFACE_MAP_ENTRY(nsIInterfaceRequestor)
   NS_INTERFACE_MAP_ENTRY(nsIWebBrowserChrome)
   NS_INTERFACE_MAP_ENTRY(nsIWebProgressListener)
   NS_INTERFACE_MAP_ENTRY(nsIEmbeddingSiteWindow)
   NS_INTERFACE_MAP_ENTRY(nsISupportsWeakReference)
NS_INTERFACE_MAP_END

//////////////////////////////////////////////
//// anyaBrowser public

anyaBrowser::anyaBrowser()
 : mLoadCount(0)
{
    NS_INIT_ISUPPORTS();

}

anyaBrowser::~anyaBrowser()
{
    nsWeakPtr weakling(
      dont_AddRef(NS_GetWeakReference((nsIWebProgressListener*)this)));
    mWebBrowser->RemoveWebBrowserListener(weakling, NS_GET_IID(nsIWebProgressListener));
}

nsresult
anyaBrowser::Init()
{
    // Initialize Gecko embedding
    anyaDirProvider *provider = new anyaDirProvider();
    NS_ADDREF(provider);
    nsresult rv = NS_InitEmbedding(nsnull, provider);
    NS_ENSURE_SUCCESS(rv, rv);

    // Use the default profile under the "Anya" directory
    rv = SelectDefaultProfile();
    NS_ENSURE_SUCCESS(rv, rv);

    mWebBrowser = do_CreateInstance(NS_WEBBROWSER_CONTRACTID, &rv);
    NS_ENSURE_SUCCESS(rv, rv);
    mWebBrowser->SetContainerWindow
	    (NS_STATIC_CAST(nsIWebBrowserChrome*, this)); 

    nsCOMPtr<nsIBaseWindow> baseWindow = do_QueryInterface(mWebBrowser);
    baseWindow->InitWindow(this, nsnull, 0, 0, 0, 0);
    rv = baseWindow->Create();
  
    // Listen to page load progress
    nsWeakPtr weakling(
      dont_AddRef(NS_GetWeakReference((nsIWebProgressListener*)this)));
    mWebBrowser->AddWebBrowserListener(weakling, NS_GET_IID(nsIWebProgressListener));

    // Create our application shell
    mAppShell = do_CreateInstance(kAppShellCID);
    NS_ENSURE_SUCCESS(rv, rv);

    return rv;
}

nsresult
anyaBrowser::SetViewportSize(nscoord aWidth, nscoord aHeight)
{
    nsCOMPtr<nsIBaseWindow> baseWindow = do_QueryInterface(mWebBrowser);
    baseWindow->SetSize(aWidth, aHeight, PR_FALSE);

    return NS_OK;
}

nsresult
anyaBrowser::GetCanvasSize(nscoord *aWidth, nscoord *aHeight)
{
    nsIFrame *canvasFrame = nsnull;
    nsresult rv = GetCanvasFrame(mWebBrowser, &canvasFrame);
    NS_ENSURE_SUCCESS(rv, rv);

    nsRect canvasRect;
    canvasFrame->GetRect(canvasRect);
    *aWidth = canvasRect.width;
    *aHeight = canvasRect.height;

    nsCOMPtr<nsIDocShell> docShell = do_GetInterface(mWebBrowser);
    nsCOMPtr<nsIPresShell> presShell;
    docShell->GetPresShell(getter_AddRefs(presShell));
    NS_ENSURE_TRUE(presShell, NS_ERROR_FAILURE);
    nsCOMPtr<nsIPresContext> presContext;
    presShell->GetPresContext(getter_AddRefs(presContext));
    float t2p;
    presContext->GetTwipsToPixels(&t2p);
    
    *aWidth = canvasRect.width * t2p;
    *aHeight = canvasRect.height * t2p;

    return NS_OK;
}

nsresult
anyaBrowser::ResizeViewportToFitCanvas()
{
    // Get the dimensions of the canvas frame
    nsIFrame *canvasFrame = nsnull;
    nsresult rv = GetCanvasFrame(mWebBrowser, &canvasFrame);
    NS_ENSURE_SUCCESS(rv, rv);
    nsRect canvasRect;
    canvasFrame->GetRect(canvasRect);

    // Get the presShell for the currently loaded document
    nsCOMPtr<nsIDocShell> docShell = do_GetInterface(mWebBrowser);
    nsCOMPtr<nsIPresShell> presShell;
    docShell->GetPresShell(getter_AddRefs(presShell));
    NS_ENSURE_TRUE(presShell, NS_ERROR_FAILURE);

    // Reflow the document with the dimensions of the canvas frame
    return presShell->ResizeReflow(canvasRect.width, canvasRect.height);
}


nsresult
anyaBrowser::LoadURI(const char *aURI, int method)
{
    mLoadCount = 0;

    nsCOMPtr<nsIInputStream> postData;

    nsString uri = NS_ConvertASCIItoUCS2(aURI);
    nsCString query;

    if (method == 1) {
        const char *q = strchr(aURI, '?');

        if (q) {
            query = q + 1;

            uri.SetLength(q - aURI);

            nsresult rv;
            nsCOMPtr<nsIInputStream> dataStream;
            rv = NS_NewCStringInputStream(getter_AddRefs(dataStream), query);
            NS_ENSURE_SUCCESS(rv, rv);

            nsCOMPtr<nsIMIMEInputStream> mimeStream(do_CreateInstance("@mozilla.org/network/mime-input-stream;1", &rv));
            NS_ENSURE_SUCCESS(rv, rv);

            mimeStream->AddHeader("Content-Type", "application/x-www-form-urlencoded");
            mimeStream->SetAddContentLength(PR_TRUE);
            mimeStream->SetData(dataStream);

            postData = mimeStream;
        }
    }

    // Tell Gecko to load the URI
    nsCOMPtr<nsIWebNavigation> webNav = do_QueryInterface(mWebBrowser);
    nsresult rv = webNav->LoadURI(uri.get(),
                                  nsIWebNavigation::LOAD_FLAGS_NONE,
                                  nsnull, postData, nsnull);
    NS_ENSURE_SUCCESS(rv, rv);

    // Wait in event loop until page load completes...
    mAppShell->Run();

    ResizeViewportToFitCanvas();

    // The page is now loaded, so we can serialize its contents
    DumpGraphics();
    DumpContent();

    return rv;
}

nsresult
anyaBrowser::DumpGraphics()
{
    // Get the scrollport widget for the currently loaded document
    nsIView *scrollPortView = nsnull;
    nsresult rv = GetScrollPortView(mWebBrowser, &scrollPortView);
    NS_ENSURE_SUCCESS(rv, rv);
    nsCOMPtr<nsIWidget> widget;
    rv = scrollPortView->GetWidget(*getter_AddRefs(widget));
    NS_ENSURE_SUCCESS(rv, rv);

    // Update the widget, which kicks off a full paint
    return widget->Update();
}

nsresult
anyaBrowser::DumpContent()
{
    // Get the currently loaded document
    nsCOMPtr<nsIDocShell> docShell = do_GetInterface(mWebBrowser);
    nsCOMPtr<nsIPresShell> presShell;
    docShell->GetPresShell(getter_AddRefs(presShell));
    NS_ENSURE_TRUE(presShell, NS_ERROR_FAILURE);
    nsCOMPtr<nsIDocument> doc;
    presShell->GetDocument(getter_AddRefs(doc));
    nsCOMPtr<nsIDOMDocument> domDoc = do_QueryInterface(doc);

    unsigned long size = 0;
    nsresult rv = DumpAnchors(domDoc, docShell, ADTM_XCODER(), &size);
    NS_ENSURE_SUCCESS(rv, rv);
    ADTM_INC(size);

    size = 0;
    rv = DumpForms(domDoc, docShell, ADTM_XCODER(), &size);
    NS_ENSURE_SUCCESS(rv, rv);
    ADTM_INC(size);

    size = 0;
    rv = DumpImageMaps(domDoc, docShell, ADTM_XCODER(), &size);
    NS_ENSURE_SUCCESS(rv, rv);
    ADTM_INC(size);

    return NS_OK;
}

//////////////////////////////////////////////
//// nsIWebBrowserChrome

NS_IMETHODIMP
anyaBrowser::SetStatus(PRUint32 aType, const PRUnichar* aStatus)
{
    return NS_OK;
}

NS_IMETHODIMP
anyaBrowser::GetWebBrowser(nsIWebBrowser** aWebBrowser)
{
    NS_ENSURE_ARG_POINTER(aWebBrowser);
    *aWebBrowser = mWebBrowser;
    NS_IF_ADDREF(*aWebBrowser);
    return NS_OK;
}

NS_IMETHODIMP
anyaBrowser::SetWebBrowser(nsIWebBrowser* aWebBrowser)
{
    mWebBrowser = aWebBrowser;
    return NS_OK;
}

NS_IMETHODIMP
anyaBrowser::GetChromeFlags(PRUint32* aChromeMask)
{
    *aChromeMask = mChromeFlags;
    return NS_OK;
}

NS_IMETHODIMP
anyaBrowser::SetChromeFlags(PRUint32 aChromeMask)
{
    mChromeFlags = aChromeMask;
    return NS_OK;
}

NS_IMETHODIMP
anyaBrowser::DestroyBrowserWindow(void)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP
anyaBrowser::SizeBrowserTo(PRInt32 aWidth, PRInt32 aHeight)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP
anyaBrowser::ShowAsModal(void)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP
anyaBrowser::IsWindowModal(PRBool *_retval)
{
    *_retval = PR_FALSE;
    return NS_OK;
}

NS_IMETHODIMP
anyaBrowser::ExitModalEventLoop(nsresult aStatus)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}

//////////////////////////////////////////////
//// nsIEmbeddingSiteWindow

NS_IMETHODIMP
anyaBrowser::SetDimensions(PRUint32 aFlags, PRInt32 x, PRInt32 y, PRInt32 cx, PRInt32 cy)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP
anyaBrowser::GetDimensions(PRUint32 aFlags, PRInt32 *x, PRInt32 *y, PRInt32 *cx, PRInt32 *cy)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP
anyaBrowser::SetFocus()
{
    return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP
anyaBrowser::GetTitle(PRUnichar * *aTitle)
{
   *aTitle = nsnull;
   return NS_OK;
}

NS_IMETHODIMP
anyaBrowser::SetTitle(const PRUnichar * aTitle)
{
    return NS_OK;
}

NS_IMETHODIMP
anyaBrowser::GetVisibility(PRBool * aVisibility)
{
    *aVisibility = PR_TRUE;
    return NS_OK;
}

NS_IMETHODIMP
anyaBrowser::SetVisibility(PRBool aVisibility)
{
    return NS_OK;
}

NS_IMETHODIMP
anyaBrowser::GetSiteWindow(void **aSiteWindow)
{
   *aSiteWindow = this;
   return NS_OK;
}

//////////////////////////////////////////////
//// nsIWebProgressListener

NS_IMETHODIMP
anyaBrowser::OnProgressChange(nsIWebProgress* aProgress,
                              nsIRequest* aRequest,
                              PRInt32 aCurSelfProgress,
                              PRInt32 aMaxSelfProgress, 
                              PRInt32 aCurTotalProgress,
                              PRInt32 aMaxTotalProgress)
{
    return NS_OK;
}

NS_IMETHODIMP
anyaBrowser::OnStateChange(nsIWebProgress* aProgress,
                           nsIRequest* aRequest,
                           PRUint32 aProgressStateFlags,
                           nsresult aStatus)
{
    // A document or sub-document is starting to load
    if ((aProgressStateFlags & STATE_IS_DOCUMENT) &&
        (aProgressStateFlags & STATE_START)) {
        ++mLoadCount;
    }

    // A document or sub-document has finished loading
    if ((aProgressStateFlags & STATE_IS_DOCUMENT) &&
        (aProgressStateFlags & STATE_STOP)) {
        --mLoadCount;

        // The document and all sub-documents are loaded, so trigger painting
        if (mLoadCount == 0)
            mAppShell->Exit();
    }

    return NS_OK;
}

NS_IMETHODIMP
anyaBrowser::OnLocationChange(nsIWebProgress* aWebProgress,
                              nsIRequest* aRequest,
                              nsIURI* aURI)
{
    return NS_OK;
}

NS_IMETHODIMP 
anyaBrowser::OnStatusChange(nsIWebProgress* aWebProgress,
                            nsIRequest* aRequest,
                            nsresult aStatus,
                            const PRUnichar* aMessage)
{
    return NS_OK;
}

NS_IMETHODIMP 
anyaBrowser::OnSecurityChange(nsIWebProgress *aWebProgress, 
                              nsIRequest *aRequest, 
                              PRUint32 state)
{
    return NS_OK;
}

//////////////////////////////////////////////
//// nsIInterfaceRequestor

NS_IMETHODIMP
anyaBrowser::GetInterface(const nsIID &aIID, void** aInstancePtr)
{
    NS_ENSURE_ARG_POINTER(aInstancePtr);

    *aInstancePtr = 0;
    if (aIID.Equals(NS_GET_IID(nsIDOMWindow)))
    {
        if (mWebBrowser)
        {
            return mWebBrowser->GetContentDOMWindow((nsIDOMWindow **) aInstancePtr);
        }
        return NS_ERROR_NOT_INITIALIZED;
    }
    return QueryInterface(aIID, aInstancePtr);
}

//////////////////////////////////////////////
//// Utilities

nsresult
SelectDefaultProfile()
{
    nsresult rv;
    nsCOMPtr<nsIProfile> profileService = 
             do_GetService(NS_PROFILE_CONTRACTID, &rv);
    NS_ENSURE_SUCCESS(rv, rv);

    PRInt32 profileCount = 0;
    rv = profileService->GetProfileCount(&profileCount);
    if (profileCount == 0)
    {
        // Make a new default profile
        NS_NAMED_LITERAL_STRING(newProfileName, "default");
        rv = profileService->CreateNewProfile(newProfileName.get(), nsnull, nsnull, PR_FALSE);
        NS_ENSURE_SUCCESS(rv, rv);
        rv = profileService->SetCurrentProfile(newProfileName.get());
        NS_ENSURE_SUCCESS(rv, rv);
    }
    else if (profileCount == 1)
    {
        // Use the current profile
        nsXPIDLString   currProfileName;
        rv = profileService->GetCurrentProfile(getter_Copies(currProfileName));
        if (NS_FAILED(rv)) return FALSE;
        rv = profileService->SetCurrentProfile(currProfileName);
        if (NS_FAILED(rv)) return FALSE;
    }

    return NS_OK;
}

nsresult
GetViewportView(nsIWebBrowser *aWebBrowser, nsIView **aView)
{
    // Get the presShell for the currently loaded document
    nsCOMPtr<nsIDocShell> docShell = do_GetInterface(aWebBrowser);
    nsCOMPtr<nsIPresShell> presShell;
    docShell->GetPresShell(getter_AddRefs(presShell));
    NS_ENSURE_TRUE(presShell, NS_ERROR_FAILURE);

    // Get the root view, or to "viewport view"
    nsCOMPtr<nsIViewManager> viewManager;
    presShell->GetViewManager(getter_AddRefs(viewManager));
    NS_ENSURE_TRUE(viewManager, NS_ERROR_FAILURE);
    nsIView *viewportView = nsnull;
    return viewManager->GetRootView(*&(*aView));
}

nsresult
GetScrollPortView(nsIWebBrowser *aWebBrowser, nsIView **aView)
{
    // Get the presShell for the currently loaded document
    nsCOMPtr<nsIDocShell> docShell = do_GetInterface(aWebBrowser);
    nsCOMPtr<nsIPresShell> presShell;
    docShell->GetPresShell(getter_AddRefs(presShell));
    NS_ENSURE_TRUE(presShell, NS_ERROR_FAILURE);

    // Get the root view, or to "viewport view"
    nsCOMPtr<nsIViewManager> viewManager;
    presShell->GetViewManager(getter_AddRefs(viewManager));
    NS_ENSURE_TRUE(viewManager, NS_ERROR_FAILURE);
    nsIView *viewportView = nsnull;
    viewManager->GetRootView(*&viewportView);
    NS_ENSURE_TRUE(viewportView, NS_ERROR_FAILURE);

    // This is kind-of a guessing game.  The scroll port view always
    // seems to be the first visible child of the viewport view
    nsIView *childView = nsnull;
    viewportView->GetFirstChild(childView);
    while (childView) {
        nsViewVisibility vis;
        childView->GetVisibility(vis);
        if (vis == nsViewVisibility_kShow) {
            *aView = childView;
            break;
        }
        childView->GetNextSibling(childView);
    }


    if (!childView) {
        *aView = nsnull;
        return NS_ERROR_FAILURE;
    }

    return NS_OK;
}

nsresult
GetCanvasFrame(nsIWebBrowser *aWebBrowser, nsIFrame **aFrame)
{
    // Get the view corresponding to the canvas frame
    nsIView *scrollPortView = nsnull;
    nsresult rv = GetScrollPortView(aWebBrowser, &scrollPortView);
    NS_ENSURE_SUCCESS(rv, rv);
    nsIView *canvasView = nsnull;
    scrollPortView->GetFirstChild(*&canvasView);
    NS_ENSURE_SUCCESS(rv, rv);
    
    // Get the dimensions of the canvas frame
    void *clientData = nsnull;
    canvasView->GetClientData(*&clientData);
    *aFrame = (nsIFrame *)clientData;
    NS_ENSURE_TRUE(*aFrame, NS_ERROR_FAILURE);
    
    return NS_OK;
}

//////////////////////////////////////////////
//// Content model serialization


static nsresult
GetFrameSize(nsIFrame *aFrame, nsRect &aRect, nsIPresContext *aPresContext)
{
    float t2p;
    aPresContext->GetTwipsToPixels(&t2p);

    aFrame->GetRect(aRect); /* XXX
                             we need to see if 'flow' is an imageframe and if it is
                             return the size of the image instead so that the 
                             anchor covers the entire image */

    nsPoint pt;
    nsIView *view;
    aFrame->GetOffsetFromView(aPresContext, pt, &view);

    aRect.MoveTo(pt);

    while (view) {
        view->GetPosition(&pt.x, &pt.y);
        aRect.MoveBy(pt.x, pt.y);
        view->GetParent(view);
    }

    aRect.x = NSTwipsToIntPixels(aRect.x, t2p);
    aRect.y = NSTwipsToIntPixels(aRect.y, t2p);
    aRect.width = NSTwipsToIntPixels(aRect.width, t2p);
    aRect.height = NSTwipsToIntPixels(aRect.height, t2p);

    return NS_OK;
}


static nsresult
GetNameAndValue(nsIFormControl *aFormControl, uint32 *aName, uint32 *aValue)
{
    nsAutoString uname, uvalue;

    nsCOMPtr<nsIDOMHTMLInputElement> inputElement(do_QueryInterface(aFormControl));
    if (inputElement) {
        inputElement->GetName(uname);
        inputElement->GetValue(uvalue);
    } else {
        nsCOMPtr<nsIDOMHTMLTextAreaElement> textareaElement(do_QueryInterface(aFormControl));
        if (textareaElement) {
            textareaElement->GetName(uname);
            textareaElement->GetValue(uvalue);
        } else {
            nsCOMPtr<nsIDOMHTMLSelectElement> selectElement(do_QueryInterface(aFormControl));
            if (selectElement) {
                selectElement->GetName(uname);
            } else {
                return NS_ERROR_FAILURE;
            }
        }
    }



    if (aName) {
        NS_ConvertUCS2toUTF8 name(uname);
        *aName = idmap_AddString(ADTM_IDMAP(), name.get(), name.Length());
    }
    if (aValue) {
        NS_ConvertUCS2toUTF8 value(uvalue);
        *aValue = idmap_AddString(ADTM_IDMAP(), value.get(), value.Length());
    }
    /* XXX we need to be able to get the raw password out of password fields */

    return NS_OK;
}

static nsresult
DumpFormChildren(nsIDOMHTMLCollection *aElements,
                 PRUint32 formid,
                 nsIPresShell *aPresShell, nsIPresContext *aPresContext,
                 nsIDOMDocument *aDomDoc, nsIDocShell* aDocShell,
                 XCoderStream *aStream, unsigned long *aSize)
{
    PRUint32 i, len;
    aElements->GetLength(&len);

    for (i=0; i < len; ++i) {
        nsCOMPtr<nsIDOMNode> node;
        aElements->Item(i, getter_AddRefs(node));

        nsCOMPtr<nsIContent> content(do_QueryInterface(node));
        nsCOMPtr<nsIFormControl> control(do_QueryInterface(node));
        if (!content || !control)
            continue;
        
        int type = control->GetType();

        switch (type) {
        case NS_FORM_INPUT_HIDDEN:
            {
                struct InputHidden hidden;
                hidden.i.formid = formid;
                GetNameAndValue(control, &hidden.name, &hidden.value);
                xcoder_InputHidden(aStream, &hidden);
                *aSize += sizeof(struct InputHidden);

                ADTM_INCFORM(1);
            }
            break;
        case NS_FORM_TEXTAREA:
        case NS_FORM_INPUT_TEXT:
        case NS_FORM_INPUT_PASSWORD:
            {
                nsIFrame *frame;
                aPresShell->GetPrimaryFrameFor(content, &frame);
                if (!frame)
                    break;

                struct InputText text;
                text.i.formid = formid;
                GetNameAndValue(control, &text.name, &text.value);

                nsRect r;
                GetFrameSize(frame, r, aPresContext);

                text.x = r.x;
                text.y = r.y;
                text.w = r.width;
                text.h = r.height;

                nsCOMPtr<nsIDOMHTMLInputElement> inputElement(do_QueryInterface(content));
                if (inputElement) {
                    PRInt32 maxLength = 0;
                    inputElement->GetMaxLength(&maxLength);
                    text.maxlength = maxLength;
                }

                uint8 abc = ABC_INPUT_TEXT;
                if (type == NS_FORM_TEXTAREA)
                    abc = ABC_INPUT_TEXTAREA;
                else if (type == NS_FORM_INPUT_PASSWORD)
                    abc = ABC_INPUT_PASSWORD;
                else
                    abc = ABC_INPUT_TEXT;
                
                xcoder_InputText(aStream, abc, &text);
                *aSize += sizeof(struct InputText);

                ADTM_INCFORM(1);
            }
            break;
        case NS_FORM_INPUT_IMAGE:
        case NS_FORM_INPUT_SUBMIT:
            {
                nsIFrame *frame;
                aPresShell->GetPrimaryFrameFor(content, &frame);
                if (!frame)
                    break;

                struct InputSubmit submit;
                submit.i.formid = formid;
                GetNameAndValue(control, &submit.name, &submit.value);

                nsRect r;
                GetFrameSize(frame, r, aPresContext);

                submit.x = r.x;
                submit.y = r.y;
                submit.w = r.width;
                submit.h = r.height;
                xcoder_InputSubmit(aStream, &submit);
                *aSize += sizeof(struct InputSubmit);

                ADTM_INCFORM(1);
            }
            break;
        case NS_FORM_INPUT_CHECKBOX:
            {
                nsIFrame *frame;
                aPresShell->GetPrimaryFrameFor(content, &frame);
                if (!frame)
                    break;

                struct InputCheckbox checkbox;
                checkbox.i.formid = formid;
                GetNameAndValue(control, &checkbox.name, NULL);

                nsCOMPtr<nsIDOMHTMLInputElement> checkboxElt = do_QueryInterface(content);
                checkboxElt->GetChecked((int *)&checkbox.checked);

                nsRect r;
                GetFrameSize(frame, r, aPresContext);

                checkbox.x = r.x;
                checkbox.y = r.y;
                checkbox.w = r.width;
                checkbox.h = r.height;
                xcoder_InputCheckbox(aStream, &checkbox);
                *aSize += sizeof(struct InputCheckbox);

                ADTM_INCFORM(1);
            }
            break;
        case NS_FORM_INPUT_RADIO:
            {
                nsIFrame *frame;
                aPresShell->GetPrimaryFrameFor(content, &frame);
                if (!frame)
                    break;

                struct InputRadio radio;
                radio.i.formid = formid;
                GetNameAndValue(control, &radio.name, &radio.value);

                nsCOMPtr<nsIDOMHTMLInputElement> radioElt = do_QueryInterface(content);
                radioElt->GetChecked((int *)&radio.checked);

                nsRect r;
                GetFrameSize(frame, r, aPresContext);

                radio.x = r.x;
                radio.y = r.y;
                radio.w = r.width;
                radio.h = r.height;
                xcoder_InputRadio(aStream, &radio);
                *aSize += sizeof(struct InputRadio);

                ADTM_INCFORM(1);
            }
            break;
        case NS_FORM_SELECT:
             {
                nsIFrame *frame;
                aPresShell->GetPrimaryFrameFor(content, &frame);
                if (!frame)
                    break;

                struct InputSelect select;
                select.i.formid = formid;
                select.selectedindex = 0;
                GetNameAndValue(control, &select.name, 0);

                nsRect r;
                GetFrameSize(frame, r, aPresContext);
                select.x = r.x;
                select.y = r.y;
                select.w = r.width;
                select.h = r.height;

                nsCOMPtr<nsIDOMHTMLSelectElement> selectElt = do_QueryInterface(content);
                
                PRInt32 size;
                selectElt->GetSize(&size);
                select.iscombo = size < 2;

                nsCOMPtr<nsIDOMHTMLCollection> options;
                selectElt->GetOptions(getter_AddRefs(options));
                if (options) {
                    PRUint32 k, len;
                    options->GetLength(&len);

                    select.noptions = len;
                    select.options = (struct InputOption *)malloc(sizeof(struct InputOption) * len);
                    for (k=0; k < len; ++k) {
                        nsCOMPtr<nsIDOMNode> optionNode;
                        options->Item(k, getter_AddRefs(optionNode));
                        nsCOMPtr<nsIDOMHTMLOptionElement> optionElt = do_QueryInterface(optionNode);

                        nsAutoString ulabel, uvalue;
                        optionElt->GetText(ulabel);
                        optionElt->GetValue(uvalue);
                        NS_ConvertUCS2toUTF8 label(ulabel);
                        NS_ConvertUCS2toUTF8 value(uvalue);

                        select.options[k].label = idmap_AddString(ADTM_IDMAP(), label.get(), label.Length());
                        select.options[k].value = idmap_AddString(ADTM_IDMAP(), value.get(), value.Length());

                        PRBool selected = PR_FALSE;
                        optionElt->GetSelected(&selected);
                        if (selected)
                            select.selectedindex = k;
                    }
                }

                xcoder_InputSelect(aStream, &select);
                *aSize += sizeof(struct InputSelect);
                ADTM_INCFORM(1);

                if (select.options)
                    free(select.options);
            }
            break;
        default:
            break;
        }
    }

    return NS_OK;
}

static nsresult
DumpImageSubmits(nsIDOMHTMLFormElement *aForm,
                 PRUint32 formid,
                 nsIPresShell *aPresShell, nsIPresContext *aPresContext,
                 XCoderStream *aStream, unsigned long *aSize)
{
    nsCOMPtr<nsIDOMNodeList> inputs;
    aForm->GetElementsByTagName(NS_LITERAL_STRING("input"), getter_AddRefs(inputs));

    PRUint32 k, len;
    inputs->GetLength(&len);
    for (k = 0; k < len; ++k) {
        nsCOMPtr<nsIDOMNode> node;
        inputs->Item(k, getter_AddRefs(node));
        nsCOMPtr<nsIContent> content(do_QueryInterface(node));
        nsCOMPtr<nsIFormControl> control(do_QueryInterface(node));
        if (!content || !control)
            continue;

        int type = control->GetType();
        if (type == NS_FORM_INPUT_IMAGE) {
            nsIFrame *frame;
            aPresShell->GetPrimaryFrameFor(content, &frame);
            if (!frame)
                continue;

            int type = control->GetType();

            struct InputSubmit submit;
            submit.i.formid = formid;
            GetNameAndValue(control, &submit.name, &submit.value);

            nsRect r;
            GetFrameSize(frame, r, aPresContext);

            submit.x = r.x;
            submit.y = r.y;
            submit.w = r.width;
            submit.h = r.height;
            xcoder_InputSubmit(aStream, &submit);
            *aSize += sizeof(struct InputSubmit);

            ADTM_INCFORM(1);
        }
    }

    return NS_OK;
}

nsresult
DumpForms(nsIDOMDocument *aDomDoc, nsIDocShell* aDocShell,
          XCoderStream *aStream, unsigned long *aSize)
{
    nsCOMPtr<nsIDOMNodeList> list;
    aDomDoc->GetElementsByTagName(NS_LITERAL_STRING("form"), getter_AddRefs(list));

    nsCOMPtr<nsIPresShell> presShell;
    aDocShell->GetPresShell(getter_AddRefs(presShell));
    if (!presShell) return NS_ERROR_FAILURE;
    nsCOMPtr<nsIPresContext> presContext;
    presShell->GetPresContext(getter_AddRefs(presContext));

    struct InputForm inform;

    PRUint32 i, len;
    list->GetLength(&len);

    for (i=0; i < len; ++i) {
        nsCOMPtr<nsIDOMNode> node;
        list->Item(i, getter_AddRefs(node));

        nsCOMPtr<nsIDOMHTMLFormElement> form(do_QueryInterface(node));
        if (!form)
            continue;

        nsCOMPtr<nsIForm> theform(do_QueryInterface(form));
        if (!theform)
            continue;

        nsCOMPtr<nsIURI> actionURI;
        theform->GetActionURL(getter_AddRefs(actionURI));
        if (!actionURI)
            continue; /* XXX make sure this isn't very bad */
        nsCAutoString str;
        actionURI->GetSpec(str);

        inform.i.formid = i;
        inform.action = idmap_AddString(ADTM_IDMAP(), str.get(), str.Length());

        inform.method = 0;
        nsAutoString methodString;
        form->GetMethod(methodString);
        if (methodString.EqualsIgnoreCase("post"))
            inform.method = 1;

        ADTM_INCFORM(1);

        xcoder_InputForm(aStream, &inform);
        *aSize += sizeof(struct InputForm);

        nsCOMPtr<nsIDOMHTMLCollection> elements;
        form->GetElements(getter_AddRefs(elements));

        DumpFormChildren(elements, i, presShell, presContext,
                         aDomDoc, aDocShell, aStream, aSize);
        DumpImageSubmits(form, i, presShell, presContext, aStream, aSize);
    }

    return NS_OK;
}

nsresult
DumpAnchors(nsIDOMDocument *aDomDoc, nsIDocShell* aDocShell,
            XCoderStream *aStream, unsigned long *aSize)
{
    nsCOMPtr<nsIDOMNodeList> list;
    aDomDoc->GetElementsByTagName(NS_LITERAL_STRING("a"), getter_AddRefs(list));

    nsCOMPtr<nsIPresShell> presShell;
    aDocShell->GetPresShell(getter_AddRefs(presShell));
    if (!presShell) return NS_ERROR_FAILURE;
    nsCOMPtr<nsIPresContext> presContext;
    presShell->GetPresContext(getter_AddRefs(presContext));

    ozRect rects[128];
    struct Event e;
    e.rects = rects;

    PRUint32 i, len;
    list->GetLength(&len);

    for (i=0; i < len; ++i) {
        nsCOMPtr<nsIDOMNode> node;
        list->Item(i, getter_AddRefs(node));

        nsCOMPtr<nsIContent> content(do_QueryInterface(node));
        if (content) {
            nsIFrame *frame;
            presShell->GetPrimaryFrameFor(content, &frame);

            if (!frame)
                continue;

            /* check to see if the frame is hidden */
            nsStyleVisibility *visibility;
            frame->GetStyleData(nsStyleVisibility::GetStyleStructID(),
                                *NS_REINTERPRET_CAST(nsStyleStruct**, &visibility));

            if (visibility->IsVisible()) {
                nsCOMPtr<nsILink> link(do_QueryInterface(content));
                if (link) {
                    nsXPIDLCString str;
                    link->GetHrefCString(*getter_Copies(str));
                    e.id = idmap_AddString(ADTM_IDMAP(), str.get(), str.Length());

                } else {
                    e.id = IDMAP_EMPTY_STRING;
                }
                nsIFrame *flow = frame;
                int n = 0;
                do {
                    nsRect r;

                    GetFrameSize(flow, r, presContext);

                    e.rects[n].x = r.x;
                    e.rects[n].y = r.y;
                    e.rects[n].width = r.width;
                    e.rects[n].height = r.height;

                    nsIFrame* firstChild = nsnull;
                    flow->FirstChild(presContext, nsnull, &firstChild);
                    if (firstChild) {
                        nsCOMPtr<nsIImageFrame> iF(do_QueryInterface(firstChild));
                        if (iF) {
                            nsCOMPtr<imgIRequest> iReq;
                            iF->GetImageRequest(getter_AddRefs(iReq));
                            nsCOMPtr<imgIContainer> iContainer;
                            iReq->GetImage(getter_AddRefs(iContainer));
                            if (!iContainer) continue;

                            int width, height;
                            iContainer->GetWidth(&width);
                            iContainer->GetHeight(&height);

                            e.rects[n].x = e.rects[n].x - (width - e.rects[n].width);
                            e.rects[n].y = e.rects[n].y - (height - e.rects[n].height);
                            e.rects[n].width = width;
                            e.rects[n].height = height;
                        }
                    }

                    ++n;
                } while (NS_SUCCEEDED(flow->GetNextInFlow(&flow)) && flow);

                e.nrects = n;
                xcoder_Event(aStream, &e);
                *aSize += sizeof(struct Event);
            }
        }
    }

    return NS_OK;
}

nsresult
DumpImageMaps(nsIDOMDocument* aDomDoc, nsIDocShell* aDocShell, 
              XCoderStream* aStream, unsigned long* aSize)
{
    *aSize = 0;

    nsCOMPtr<nsIDOMNodeList> images;
    aDomDoc->GetElementsByTagName(NS_LITERAL_STRING("img"), getter_AddRefs(images));

    PRUint32 imageCount;
    images->GetLength(&imageCount);

    for (int i = 0; i < imageCount; ++i) {
        nsCOMPtr<nsIDOMNode> imgNode;
        images->Item(i, getter_AddRefs(imgNode));

        nsCOMPtr<nsIDOMElement> imageElt(do_QueryInterface(imgNode));

        PRBool hasUsemapAttr = PR_FALSE;
        imageElt->HasAttribute(NS_LITERAL_STRING("usemap"), &hasUsemapAttr);
        if (hasUsemapAttr) {
            nsAutoString usemap;
            imageElt->GetAttribute(NS_LITERAL_STRING("usemap"), usemap);

            nsCOMPtr<nsIDOMHTMLMapElement> mapElt;
            nsCOMPtr<nsIDocument> doc(do_QueryInterface(aDomDoc));
            nsImageMapUtils::FindImageMap(doc, usemap, getter_AddRefs(mapElt));

            if (mapElt) {
                nsCOMPtr<nsIDOMHTMLCollection> areas;
                mapElt->GetAreas(getter_AddRefs(areas));

                PRUint32 areaCount;
                areas->GetLength(&areaCount);

                for (int j = 0; j < areaCount; ++j) {
                    nsCOMPtr<nsIDOMNode> areaNode;
                    areas->Item(j, getter_AddRefs(areaNode));

                    nsCOMPtr<nsIDOMElement> areaElt(do_QueryInterface(areaNode));

                    PRBool hasNoHrefAttr = PR_FALSE;
                    areaElt->HasAttribute(NS_LITERAL_STRING("nohref"), &hasNoHrefAttr);
                    if (hasNoHrefAttr)
                        break;

                    nsAutoString coords;
                    areaElt->GetAttribute(NS_LITERAL_STRING("coords"), coords);

                    nsAutoString shape;
                    
                    ozRect boundingBox;
                    // Bounding box is initially the image itself (coords are specified relative
                    // to the position of the image, and also the 'default' shape specifies the 
                    // entire image so this is useful to us.
                    GetImageRect(imageElt, aDocShell, &boundingBox);

                    areaElt->GetAttribute(NS_LITERAL_STRING("shape"), shape);
                    if (shape.EqualsIgnoreCase("rect"))
                        FindBoundingBoxForRectArea(coords, &boundingBox);
                    else if (shape.EqualsIgnoreCase("poly"))
                        FindBoundingBoxForPolyArea(coords, &boundingBox);
                    else if (shape.EqualsIgnoreCase("circle"))
                        FindBoundingBoxForCircleArea(coords, &boundingBox);

                    ozRect rects[1];
                    struct Event e;
                    e.rects = rects;
                    e.rects[0] = boundingBox;
                    e.nrects = 1;

                    nsAutoString href;
                    areaElt->GetAttribute(NS_LITERAL_STRING("href"), href);
                    
                    char* s = ToNewCString(href);
                    if (s) {
                        e.id = idmap_AddString(ADTM_IDMAP(), s, href.Length());
                        nsCRT::free(s);
                    }
                
                    xcoder_Event(aStream, &e);
                    *aSize += sizeof(struct Event);
                }
            }
        }
    }

    return NS_OK;
}

nsresult
GetImageRect(nsIDOMElement* aImageElement, nsIDocShell* aDocShell, ozRect* aImageRect)
{
    aImageRect->x = 0; aImageRect->y = 0; aImageRect->width = 0; aImageRect->height = 0;

    nsCOMPtr<nsIContent> content(do_QueryInterface(aImageElement));
    if (content) {
        nsCOMPtr<nsIPresShell> presShell;
        aDocShell->GetPresShell(getter_AddRefs(presShell));
        if (!presShell) return NS_ERROR_FAILURE;
        nsCOMPtr<nsIPresContext> presContext;
        presShell->GetPresContext(getter_AddRefs(presContext));

        nsIFrame* frame;
        presShell->GetPrimaryFrameFor(content, &frame);

        if (!frame)
            return NS_OK;

        nsStyleVisibility* visibility;
        frame->GetStyleData(nsStyleVisibility::GetStyleStructID(), 
                            *NS_REINTERPRET_CAST(nsStyleStruct**, &visibility));

        if (visibility->IsVisible()) {
            nsIView* view;
            float t2p;
            nsRect r;
            frame->GetRect(r);
            nsPoint pt;
            frame->GetOffsetFromView(presContext, pt, &view);

            r.MoveTo(pt);

            while (view) {
                view->GetPosition(&pt.x, &pt.y);
                r.MoveBy(pt.x, pt.y);
                view->GetParent(view);
            }

            presContext->GetTwipsToPixels(&t2p);
            r.x = NSTwipsToIntPixels(r.x, t2p);
            r.y = NSTwipsToIntPixels(r.y, t2p);
            r.width = NSTwipsToIntPixels(r.width, t2p);
            r.height = NSTwipsToIntPixels(r.height, t2p);

            nsCOMPtr<nsIImageFrame> imgFrame(do_QueryInterface(frame));
            if (imgFrame) {
                nsCOMPtr<imgIRequest> req;
                imgFrame->GetImageRequest(getter_AddRefs(req));
                if (req) {
                    nsCOMPtr<imgIContainer> container;
                    req->GetImage(getter_AddRefs(container));
                    if (container) {
                        nsCOMPtr<gfxIImageFrame> gif;
                        container->GetCurrentFrame(getter_AddRefs(gif));
                        if (gif) {
                            nsCOMPtr<nsIImage> img(do_GetInterface(gif));
                            if (img) {
                                r.width = img->GetWidth();
                                r.height = img->GetHeight();
                            }
                        }
                    }
                }
            }

            aImageRect->x = r.x;
            aImageRect->y = r.y;
            aImageRect->width = r.width;
            aImageRect->height = r.height;
        }
    }

    return NS_OK;
}

void
FindBoundingBoxForRectArea(const nsAString& aCoords, ozRect* aBoundingBox)
{
    PRInt32* coords = nsnull;
    PRInt32 numCoords = 0;
    char* cp = ToNewCString(aCoords);
    if (cp) {
        coords = lo_parse_coord_list(cp, &numCoords);
        nsCRT::free(cp);
    }    

    aBoundingBox->x += coords[0];
    aBoundingBox->y += coords[1];
    aBoundingBox->width = coords[2] - coords[0];
    aBoundingBox->height = coords[3] - coords[1];

    delete[] coords;
}

void 
FindBoundingBoxForPolyArea(const nsAString& aCoords, ozRect* aBoundingBox)
{
    PRInt32* coords = nsnull;
    PRInt32 numCoords = 0;
    char* cp = ToNewCString(aCoords);
    if (cp) {
        coords = lo_parse_coord_list(cp, &numCoords);
        nsCRT::free(cp);
    }    

    PRInt32 minX, maxX, minY, maxY;
    minX = maxX = minY = maxY = 0;
    for (int i = 0; i < numCoords; ++i) {
        if (i % 2 == 0) {
            if (coords[i] < minX) 
                minX = coords[i];
            else if (coords[i] > maxX) 
                maxX = coords[i];
        }
        else {
            if (coords[i] < minY)
                minY = coords[i];
            else if (coords[i] > maxY)
                maxY = coords[i];
        }
    }

    aBoundingBox->x += minX;
    aBoundingBox->y += minY;
    aBoundingBox->width = maxX - minX;
    aBoundingBox->height = maxY - minY;

    delete[] coords;
}

void 
FindBoundingBoxForCircleArea(const nsAString& aCoords, ozRect* aBoundingBox)
{
    PRInt32* coords = nsnull;
    PRInt32 numCoords = 0;
    char* cp = ToNewCString(aCoords);
    if (cp) {
        coords = lo_parse_coord_list(cp, &numCoords);
        nsCRT::free(cp);
    }    

    aBoundingBox->x += coords[0] - (coords[2] / 2);
    aBoundingBox->y += coords[1] - (coords[2] / 2);
    aBoundingBox->width = 2 * coords[2];
    aBoundingBox->height = 2 * coords[2];

    delete[] coords;
}

inline PRBool
is_space(char c)
{
  return (c == ' ' ||
          c == '\f' ||
          c == '\n' ||
          c == '\r' ||
          c == '\t' ||
          c == '\v');
}

// XXX straight copy from laymap.c
nscoord* 
lo_parse_coord_list(char *str, PRInt32* value_cnt)
{
  char *tptr;
  char *n_str;
  PRInt32 i, cnt;
  PRInt32 *value_list;

  /*
   * Nothing in an empty list
   */
  *value_cnt = 0;
  if (!str || *str == '\0')
  {
    return nsnull;
  }

  /*
   * Skip beginning whitespace, all whitespace is empty list.
   */
  n_str = str;
  while (is_space(*n_str))
  {
    n_str++;
  }
  if (*n_str == '\0')
  {
    return nsnull;
  }

  /*
   * Make a pass where any two numbers separated by just whitespace
   * are given a comma separator.  Count entries while passing.
   */
  cnt = 0;
  while (*n_str != '\0')
  {
    PRBool has_comma;

    /*
     * Skip to a separator
     */
    tptr = n_str;
    while (!is_space(*tptr) && *tptr != ',' && *tptr != '\0')
    {
      tptr++;
    }
    n_str = tptr;

    /*
     * If no more entries, break out here
     */
    if (*n_str == '\0')
    {
      break;
    }

    /*
     * Skip to the end of the separator, noting if we have a
     * comma.
     */
    has_comma = PR_FALSE;
    while (is_space(*tptr) || *tptr == ',')
    {
      if (*tptr == ',')
      {
        if (has_comma == PR_FALSE)
        {
          has_comma = PR_TRUE;
        }
        else
        {
          break;
        }
      }
      tptr++;
    }
    /*
     * If this was trailing whitespace we skipped, we are done.
     */
    if ((*tptr == '\0')&&(has_comma == PR_FALSE))
    {
      break;
    }
    /*
     * Else if the separator is all whitespace, and this is not the
     * end of the string, add a comma to the separator.
     */
    else if (has_comma == PR_FALSE)
    {
      *n_str = ',';
    }

    /*
     * count the entry skipped.
     */
    cnt++;

    n_str = tptr;
  }
  /*
   * count the last entry in the list.
   */
  cnt++;
 
  /*
   * Allocate space for the coordinate array.
   */
  value_list = new nscoord[cnt];
  if (!value_list)
  {
    return nsnull;
  }

  /*
   * Second pass to copy integer values into list.
   */
  tptr = str;
  for (i=0; i<cnt; i++)
  {
    char *ptr;

    ptr = strchr(tptr, ',');
    if (ptr)
    {
      *ptr = '\0';
    }
    /*
     * Strip whitespace in front of number because I don't
     * trust atoi to do it on all platforms.
     */
    while (is_space(*tptr))
    {
      tptr++;
    }
    if (*tptr == '\0')
    {
      value_list[i] = 0;
    }
    else
    {
      value_list[i] = (nscoord) ::atoi(tptr);
    }
    if (ptr)
    {
      *ptr = ',';
      tptr = ptr + 1;
    }
  }

  *value_cnt = cnt;
  return value_list;
}

