/* -*- Mode: C++; tab-width: 20; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Anya server code
 *
 * The Initial Developer of the Original Code is Netscape Communications
 * Corporation. Portions created by the Initial Developer are
 * Copyright (C) 2003 the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *    Stuart Parmenter <pavlov@netscape.com>
 *    Joe Hewitt <hewitt@netscape.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 */

#include "anya.h"
#include "xcoder.h"
#include "prims.h"

#include "anyaBrowser.h"

#include "nsCOMPtr.h"
#include "nsIServiceManager.h"
#include "anyaIDataTransferManager.h"
#include "prinrval.h"

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <winsock2.h>

#include <io.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

typedef struct Channel {
    XCoderStream    stream;
} Channel;

#define NCHAN_LIMIT     20

static anyaBrowser *gBrowser;

int
read_request(XCoderStream *stream, struct Request *r)
{
    if (!xcoder_Request(stream, r)) {
        printf("can't decode Device info from client\n");
        return 0;
    }
    printf("Device info:\n");
    printf("  version: %d\n", r->version);
    printf("  display: %d\n", r->device.display);
    printf("  dpi:     %d\n", r->device.dpi);
    printf("  xres:    %d\n", r->device.xres);
    printf("  yres:    %d\n", r->device.yres);
    printf("  uid      %d\n", r->device.uid);

    printf("Request received, type %d\n", r->request);
    printf("URI: %s\n", r->uri);

    nsCOMPtr<anyaIDataTransferManager> adtm(do_GetService(NS_ANYADATAXFERMGR_CONTRACTID));
    adtm->SetRequest(r);

    return 1;
}

int
send_error(XCoderStream *stream, int send_header)
{
    uint8 error = ABC_ERROR;
    if (send_header) {
        struct Header header;
        memset(&header, 0, sizeof(struct Header));

        /* header.version == 0 indicates an error occured */
        xcoder_Header(stream, &header);
    }

    xcoder_Read8(stream, &error);
    xcoder_FlushBuffer(stream);
    xcoder_FlushBufferEnd(stream);
    xcoder_Close(stream);

    return 1;
}

int
send_data(XCoderStream *stream, const struct Request *sd)
{
    XCoderStream writer;
    int fd;
    FILE *file;
    struct Header header;

    if (sd->version != ANYA_VERSION) {
        send_error(&writer, TRUE);
        return 0;
    }

    xcoder_Open(&writer, stream->fd, XCODER_WRITE /*| XCODER_COMPRESS*/);

    /* Resize the embedded browser to match the device */
    gBrowser->SetViewportSize(sd->device.xres, sd->device.yres);

    /* Load the URI synchronously */
    nsresult rv = gBrowser->LoadURI(sd->uri, (sd->request == ANYA_REQUEST_POST));

    if (NS_FAILED(rv)) {
        send_error(&writer, TRUE);
        return 0;
    }

    /* Flush the rendering data to disk */
    nsCOMPtr<anyaIDataTransferManager> adtm;
    adtm = do_GetService(NS_ANYADATAXFERMGR_CONTRACTID);
    adtm->Flush();

    nscoord w = 0, h = 0;
    gBrowser->GetCanvasSize(&w, &h);

    header.version = ANYA_VERSION;
    header.w = w;
    header.h = h;

    /* XXX this should use a unique file name and match up with that from the ADTM */

    file = fopen("c:\\anya_len.txt", "r");
    fscanf(file, "%d %d", &header.display_list_size, &header.form_list_length);
    fclose(file);

    xcoder_Header(&writer, &header);
    xcoder_FlushBuffer(&writer);

    fd = _open("c:\\anya.data", _O_RDONLY | _O_BINARY, _S_IREAD);
    while (writer.cursor = _read(fd, &writer.buffer, XCODER_BUFFER_SIZE)) {
        xcoder_FlushBuffer(&writer);
    }
    _close(fd);
    xcoder_FlushBuffer(&writer);

    xcoder_FlushBufferEnd(&writer);

    if (writer.zstream)
        printf("%d -> %d\n", writer.zstream->total_in, writer.zstream->total_out);

    xcoder_Close(&writer);

    return 1;
}


int
handle_client(XCoderStream *stream)
{
    struct Request r;
    read_request(stream, &r);
    send_data(stream, &r);
    return 1;
}

int
main()
{
    int sock, nfd, nchan, i;
    struct sockaddr_in local_addr;
    fd_set readfds;
    Channel chan[NCHAN_LIMIT];

    WSADATA wsaData;
    if (WSAStartup(MAKEWORD(2,0), &wsaData) != 0) {
        return 0;
    }

    /* Initialize the embedded browser */
    gBrowser = new anyaBrowser();
    NS_ADDREF(gBrowser);
    nsresult rv = gBrowser->Init();
    NS_ENSURE_SUCCESS(rv, 0);

    printf("color: %d\n", sizeof(struct Color));
    printf("line:  %d\n", sizeof(struct Line));
    printf("rect:  %d\n", sizeof(struct Rect));
    printf("text:  %d\n", sizeof(struct Text));
    printf("event: %d\n", sizeof(struct Event));
    printf("image: %d\n", sizeof(struct Image));

    sock = socket(AF_INET, SOCK_STREAM, IPPROTO_IP);
    if (sock < 0) {
        printf("socket returned an error\n");
        return -1;
    }

    /* setup the local addr */
    memset(&local_addr, 0, sizeof(local_addr));
    local_addr.sin_family = AF_INET;
    local_addr.sin_port = htons(ANYA_PORT);
    local_addr.sin_addr.s_addr = htonl(INADDR_ANY);

    if (bind(sock, (struct sockaddr*)&local_addr, sizeof(local_addr)) < 0) {
        printf("bind failed\n");
        return -1;
    }

    if (listen(sock, 5) < 0) {
        printf("listen failed\n");
        return -1;
    }

    FD_ZERO(&readfds);
    FD_SET(sock, &readfds);
    nfd = sock + 1;
    nchan = 0;

    while (select(nfd, &readfds, NULL, NULL, NULL) >= 0) {
        if (FD_ISSET(sock, &readfds)) {
            struct sockaddr_in caddr;
            int caddr_len = sizeof(caddr);
            int clientfd;

            clientfd = accept(sock, (struct sockaddr*)&caddr, &caddr_len);
            if (clientfd > 0) {
                if (nchan == NCHAN_LIMIT) {
                    printf("too many channels\n");
                    return -1;
                }

                i = nchan++;
                if (!xcoder_Open(&chan[i].stream, clientfd, XCODER_READ)) {
                    printf("can't open transcoder stream\n");
                    return -1;
                }

                FD_SET(chan[i].stream.fd, &readfds);
            }
        } else {
            for (i = 0; i < nchan; i++) {
                /* XXX hack around blocking i/o for now; if we use non-blocking
                       i/o later, we need to copy what's available into a buffer
                       and process it when it fills to a certain point. */

                if (FD_ISSET(chan[i].stream.fd, &readfds) && !xcoder_FillBuffer(&chan[i].stream)) {
                    /* Can't read, error or end of connection. */
                    FD_CLR(chan[i].stream.fd, &readfds);
                    closesocket(chan[i].stream.fd);
                    xcoder_Close(&chan[i].stream);
                    if (i < nchan - 1) {
                        chan[i] = chan[nchan - 1];
                    }
                    --nchan;
                    continue;
                }

                PRIntervalTime now = PR_IntervalNow();
                handle_client(&chan[i].stream);

                printf("took %d\n", PR_IntervalToMilliseconds(PR_IntervalNow() - now));
            }

            FD_SET(sock, &readfds);
        }
    }

    closesocket(sock);

    NS_RELEASE(gBrowser);

    WSACleanup();

    return 1;
}
