/* -*- Mode: C++; tab-width: 20; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Anya client/server code
 *
 * The Initial Developer of the Original Code is Netscape Communications
 * Corporation. Portions created by the Initial Developer are
 * Copyright (C) 2003 the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *    Stuart Parmenter <pavlov@netscape.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 */

/*
 * idmap - compressed string goodness
 *
 * The idea here is this:
 *
 * Call idmap_AddString() and have it add a string to the idmap and return
 * an id that can be used to retrieve the string later from a compressed or
 * uncompressed idmap.
 *
 * After all the strings you want to be in this idmap are added, call
 * idmap_Compress() to turn the idmap in to a readonly compressed map.
 * The result of this is probably that a dictionary is built and stored
 * and the strings are compressed.  The dictionary + compressed strings
 * clearly needs to provide some signifigant savings (at least 30-40%?)
 * over the uncompressed strings.  The unique id can't be based on any
 * memory addresses since this idmap needs to be able to be transfered
 * over the network.
 *
 * Calls to idmap_GetString() on a compressed idmap needs to be able to
 * retrieve the string and its length based on the id provided.  It needs
 * to be able to do this quickly and without decompressing the idmap.
 *
 *
 * XXX The current implementation doesn't do any compression.
**/

#include "idmap.h"

#include <string.h>

#if defined (_WIN32) || defined(_WIN32_WCE)
#include <malloc.h>

#ifdef _WIN32_WCE
#include <winsock.h>
#else
#include <winsock2.h> /* for ntohl and htonl */
#endif
#endif

#ifdef PALMOS_SDK_VERSION
#define htons(x) NetHToNS(x)
#define htonl(x) NetHToNL(x)
#define ntohs(x) NetNToHL(x)
#define ntohl(x) NetNToHS(x)
#endif


void
idmap_Init(struct IdMap *map)
{
    map->map = NULL;
    map->cursor = map->count = 0;
}

void
idmap_Destroy(struct IdMap *map)
{
    if (map->map)
        free(map->map);
    map->map = NULL;
    map->cursor = map->count = 0;
}

void
idmap_Compress(struct IdMap *map)
{
    uint8 *buf;

    if (map->cursor == map->count)
        return;

    buf = (uint8 *)malloc(map->cursor);
    memcpy(buf, map->map, map->cursor);
    free(map->map);
    map->map = buf;
    map->count = map->cursor;
}

uint32
idmap_AddString(struct IdMap *map, const char *string, uint32 length)
{
    uint32 nlen;
    if (string == NULL || length == 0)
        return IDMAP_EMPTY_STRING;

    if (map->count - map->cursor <= length + (uint32)sizeof(length)) {
        int newsize = (map->cursor + (sizeof(length) + length)) * 2;
        uint8 *buf = (uint8 *)malloc(newsize);
        if (!buf)
            return 0;
        if (map->map) {
            memcpy(buf, map->map, map->cursor);
            free(map->map);
        }
        map->map = buf;
        map->count = newsize;
    }
    nlen = htonl(length);
    memcpy(map->map + map->cursor, &nlen, sizeof(nlen));
    map->cursor += sizeof(nlen);

    memcpy(map->map + map->cursor, string, length);
    map->cursor += length;

    return map->cursor - length - sizeof(nlen);
}

const char *
idmap_GetString(const struct IdMap *map, uint32 id, uint32 *length)
{
    uint32 nlen;
    if (id == IDMAP_EMPTY_STRING || !map->map) {
        if (length)
            *length = 0;
        return NULL;
    }
    memcpy(&nlen, map->map + id, sizeof(nlen));
    if (length)
        *length = ntohl(nlen);
    return (char *)map->map + id + sizeof(nlen);
}

char *
idmap_GetStringCopy(const struct IdMap *map, uint32 id, uint32 *length)
{
    char *string = NULL;
    uint32 len = 0;
    if (id != IDMAP_EMPTY_STRING || !map->map) {
        memcpy(&len, map->map + id, sizeof(len));
        len = ntohl(len);

        string = (char *)malloc(len + 1);
        if (string) {
            memcpy(string, map->map + id + sizeof(len), len);
            string[len] = '\0';
        } else {
            len = 0;
        }
    }
    if (length)
        *length = len;
    return string;
}


/* test */
#ifdef STANDALONE
int
main()
{
    struct IdMap idmap;

    static const char text[] = "The giant squid is the over there";

    uint16 id;
    char *str;
    uint16 len;

    
    idmap_Init(&idmap);

    id = idmap_AddString(&idmap, text, sizeof(text));

    idmap_Compress(&idmap);

    str = idmap_GetString(&idmap, id, &len);
    printf("%s\n", str);

    idmap_Destroy(&idmap);

    return 1;
}
#endif
