/* -*- Mode: C++; tab-width: 20; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Anya client/server code
 *
 * The Initial Developer of the Original Code is Netscape Communications
 * Corporation. Portions created by the Initial Developer are
 * Copyright (C) 2003 the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *    Stuart Parmenter <pavlov@netscape.com>
 *    Joe Hewitt <hewitt@netscape.com>
 *    Ben Googer <ben@netscape.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 */

/*
 * Anya Primative helpers
 */

#include "prims.h"
#include "idmap.h"
#include "abc.h"

#include <stdio.h>

static int
xcoder_WH(XCoderStream *stream, uint16 *w, uint16 *h)
{
    return xcoder_Read16(stream, w) &&
           xcoder_Read16(stream, h);
}

static int
xcoder_XY(XCoderStream *stream, uint16 *x, uint16 *y)
{
    return xcoder_Read16(stream, x) &&
           xcoder_Read16(stream, y);
}

int
xcoder_Request(XCoderStream *stream, struct Request *req)
{
    return xcoder_Read8 (stream, &req->version) &&
           xcoder_Read8 (stream, &req->request) &&
           xcoder_Read8 (stream, &req->device.display) &&
           xcoder_Read8 (stream, &req->device.dpi) &&
           xcoder_Read16(stream, &req->device.xres) &&
           xcoder_Read16(stream, &req->device.yres) &&
           xcoder_Read32(stream, &req->device.uid) &&
           xcoder_ReadString(stream, &req->uri);
}

int
xcoder_Header(XCoderStream *stream, struct Header *head)
{
    return xcoder_Read8(stream, &head->version) &&
           xcoder_WH(stream, &head->w, &head->h) &&
           xcoder_Read32(stream, &head->display_list_size) &&
           xcoder_Read32(stream, &head->form_list_length);
}

static int
xcoder_ABC(XCoderStream *stream, uint8 abc, uint8 *a)
{
    if (FLAG_WRITE(stream) || FLAG_WRITEFILE(stream)) {
        if (!xcoder_Read8(stream, &abc))
            return 0;
    } else if (FLAG_READ(stream)) {
        *a = abc;
    }
    return 1;
}

/* idmap */

int
xcoder_IdMap(XCoderStream *stream, struct IdMap *map)
{
    uint8 type;
    if (!xcoder_ABC(stream, ABC_IDMAP, &type) ||
        !xcoder_ReadBytes(stream, &map->cursor, &map->map))
        return 0;

    map->count = map->cursor;

    return 1;
}

/* 2d prims */

int
xcoder_Point(XCoderStream *stream, struct Point *p)
{
    return xcoder_ABC(stream, ABC_POINT, &p->type) &&
           xcoder_XY(stream, &p->x, &p->y);
}

int
xcoder_Line(XCoderStream *stream, struct Line *l)
{
    return xcoder_ABC(stream, ABC_LINE, &l->type) &&
           xcoder_XY(stream, &l->x1, &l->y1) && 
           xcoder_XY(stream, &l->x2, &l->y2);
}

int
xcoder_Rect(XCoderStream *stream, struct Rect *r)
{
    return xcoder_ABC(stream, ABC_RECT, &r->type) &&
           xcoder_XY(stream, &r->x, &r->y) && 
           xcoder_WH(stream, &r->w, &r->h) &&
           xcoder_Read8(stream, &r->filled);
}

int
xcoder_Arc(XCoderStream *stream, struct Arc *a)
{
    return xcoder_ABC(stream, ABC_ARC, &a->type) &&
           xcoder_XY(stream,  &a->x1, &a->y1) &&
           xcoder_XY(stream,  &a->x2, &a->y2) &&
           xcoder_Read16(stream, &a->start) &&
           xcoder_Read16(stream, &a->end);
}

/* handles points, polygons and polylines */
static int
xcoder_Multi(XCoderStream *stream, uint16 *npoints, ozPoint **points)
{
    ozPoint *pts = *points;
    uint16 npts = *npoints;
    uint16 i;

    if (!xcoder_Read16(stream, &npts))
        return 0;

    if (FLAG_READ(stream) && npts > 0) {
        pts = malloc(sizeof(ozPoint) * npts);
        if (!pts)
            return 0;
    }

    for (i=0; i < npts; ++i) {
        uint16 x = pts[i].x;
        uint16 y = pts[i].y;
        if (!xcoder_XY(stream, &x, &y)) {
            if (FLAG_READ(stream))
                free(pts);
            return 0;
        }
        pts[i].x = x;
        pts[i].y = y;
    }

    *npoints = npts;
    *points = pts;

    return 1;
}

int
xcoder_Points(XCoderStream *stream, struct Points *p)
{
    return xcoder_ABC(stream, ABC_POINTS, &p->type) &&
           xcoder_Multi(stream, &p->npoints, &p->points);
}

int
xcoder_Polyline(XCoderStream *stream, struct Polyline *p)
{
    return xcoder_ABC(stream, ABC_POLYLINE, &p->type) &&
           xcoder_Multi(stream, &p->npoints, &p->points);
}

int
xcoder_Polygon(XCoderStream *stream, struct Polygon *p)
{
    return xcoder_ABC(stream, ABC_POLYGON, &p->type) &&
           xcoder_Multi(stream, &p->npoints, &p->points) &&
           xcoder_Read8(stream, &p->filled);
}


int
xcoder_Text(XCoderStream *stream, struct Text *t)
{
    return xcoder_ABC(stream, ABC_TEXT, &t->type) &&
           xcoder_XY(stream, &t->x, &t->y) &&
           xcoder_Read32(stream, &t->id);
}

int
xcoder_Font(XCoderStream *stream, struct Font *f)
{
    return xcoder_ABC(stream, ABC_FONT, &f->type) &&
           xcoder_Read8(stream, &f->id);

}

int
xcoder_Color(XCoderStream *stream, struct Color *c)
{
    return xcoder_ABC(stream, ABC_COLOR, &c->type) &&
           xcoder_Read8(stream, &c->r) &&
           xcoder_Read8(stream, &c->g) &&
           xcoder_Read8(stream, &c->b);

}

int
xcoder_Image(XCoderStream *stream, struct Image *i)
{
    return xcoder_ABC(stream, ABC_IMAGE, &i->type) &&
           xcoder_Read8(stream, &i->bpp) &&
           xcoder_Read8(stream, &i->istile) &&
           xcoder_XY(stream, &i->x, &i->y) && 
           xcoder_WH(stream, &i->w, &i->h) &&
           xcoder_WH(stream, &i->destw, &i->desth) &&
           xcoder_XY(stream, &i->xoffset, &i->yoffset) && 
           xcoder_ReadBytes(stream, &i->nbytes, &i->bytes);
}

int
xcoder_Event(XCoderStream *stream, struct Event *e)
{
    uint16 i;

    if (!xcoder_ABC(stream, ABC_EVENT, &e->type) ||
        !xcoder_Read32(stream, &e->id) ||
        !xcoder_Read16(stream, &e->nrects))
        return 0;

    if (FLAG_READ(stream) && e->nrects > 0) {
        e->rects = malloc(sizeof(ozRect) * e->nrects);
        if (!e->rects)
            return 0;
    }

    for (i=0; i < e->nrects; ++i) {
        uint16 x = e->rects[i].x;
        uint16 y = e->rects[i].y;
        uint16 width = e->rects[i].width;
        uint16 height = e->rects[i].height;
        
        if (!xcoder_XY(stream, &x, &y) ||
            !xcoder_WH(stream, &width, &height)) {
            if (FLAG_READ(stream))
                free(e->rects);
            return 0;
        }
        
        e->rects[i].x = x;
        e->rects[i].y = y;
        e->rects[i].width = width;
        e->rects[i].height = height;
    }

    return 1;
}


int
xcoder_InputForm(XCoderStream *stream, struct InputForm *i)
{
    return xcoder_ABC(stream, ABC_INPUT_FORM, &i->i.type) &&
           xcoder_Read8(stream, &i->i.formid) &&
           xcoder_Read8(stream, &i->method) &&
           xcoder_Read32(stream, &i->action);
}

int
xcoder_InputHidden(XCoderStream *stream, struct InputHidden *i)
{
    return xcoder_ABC(stream, ABC_INPUT_HIDDEN, &i->i.type) &&
           xcoder_Read8(stream, &i->i.formid) &&
           xcoder_Read32(stream, &i->name) &&
           xcoder_Read32(stream, &i->value);
}

int
xcoder_InputText(XCoderStream *stream, uint8 abc, struct InputText *i)
{
    i->data = NULL;

    return xcoder_ABC(stream, abc, &i->i.type) &&
           xcoder_Read8(stream, &i->i.formid) &&
           xcoder_XY(stream, &i->x, &i->y) &&
           xcoder_WH(stream, &i->w, &i->h) &&
           xcoder_Read32(stream, &i->name) &&
           xcoder_Read32(stream, &i->value) &&
           xcoder_Read32(stream, &i->maxlength);
}

int
xcoder_InputCheckbox(XCoderStream *stream, struct InputCheckbox *i)
{
    i->data = NULL;

    return xcoder_ABC(stream, ABC_INPUT_CHECKBOX, &i->i.type) &&
           xcoder_Read8(stream, &i->i.formid) &&
           xcoder_XY(stream, &i->x, &i->y) &&
           xcoder_WH(stream, &i->w, &i->h) &&
           xcoder_Read32(stream, &i->name) &&
           xcoder_Read8(stream, &i->checked);
}

int
xcoder_InputRadio(XCoderStream *stream, struct InputRadio *i)
{
    i->data = NULL;

    return xcoder_ABC(stream, ABC_INPUT_RADIO, &i->i.type) &&
           xcoder_Read8(stream, &i->i.formid) &&
           xcoder_XY(stream, &i->x, &i->y) &&
           xcoder_WH(stream, &i->w, &i->h) &&
           xcoder_Read32(stream, &i->name) &&
           xcoder_Read32(stream, &i->value) &&
           xcoder_Read8(stream, &i->checked);
}

int
xcoder_InputSubmit(XCoderStream *stream, struct InputSubmit *i)
{
    return xcoder_ABC(stream, ABC_INPUT_SUBMIT, &i->i.type) &&
           xcoder_Read8(stream, &i->i.formid) &&
           xcoder_XY(stream, &i->x, &i->y) &&
           xcoder_WH(stream, &i->w, &i->h) &&
           xcoder_Read32(stream, &i->name) &&
           xcoder_Read32(stream, &i->value);
}

int
xcoder_InputSelect(XCoderStream *stream, struct InputSelect *i)
{
    uint32 k;

    i->data = NULL;

    if (!xcoder_ABC(stream, ABC_INPUT_SELECT, &i->i.type) ||
        !xcoder_Read8(stream, &i->i.formid) ||
        !xcoder_XY(stream, &i->x, &i->y) ||
        !xcoder_WH(stream, &i->w, &i->h) ||
        !xcoder_Read32(stream, &i->name) ||
        !xcoder_Read32(stream, &i->selectedindex) ||
        !xcoder_Read8(stream, &i->iscombo) ||
        !xcoder_Read32(stream, &i->noptions)) {
        return 0;
    }

    if (FLAG_READ(stream) && i->noptions > 0) {
        i->options = malloc(sizeof(struct InputOption) * i->noptions);
        if (!i->options)
            return 0;
    }

    for (k=0; k < i->noptions; ++k) {
        if (!xcoder_Read32(stream, &(i->options[k]).label) ||
            !xcoder_Read32(stream, &(i->options[k]).value)) {
            if (FLAG_READ(stream))
                free(i->options);
            return 0;
        }
    }

  return 1;
}


#if 0

int
xcoder_HtmlElement(XCoderStream *stream, HtmlElement *element)
{
    uint32 i;
    element->type = ABC_HTMLELEMENT; /* XXX hack */

    if (!xcoder_Read8(stream, &element->tp) ||
        !xcoder_Read32(stream, &element->nattrs))
        return 0;

    if (FLAG_READ(stream) && element->nattrs > 0)
        element->attrs = malloc(sizeof(struct _attr) * element->nattrs);

    for (i = 0; i < element->nattrs; ++i) {
        if (!xcoder_Read32(stream, &element->attrs[i].id) || 
            !xcoder_Read32(stream, &element->attrs[i].value))
            return 0;

        if (element->attrs[i].value & 1) {
            uint32 n, num = element->attrs[i].value >> 1;

            if (FLAG_READ(stream) && num > 0) {
                element->attrs[i].values = malloc(sizeof(struct _attr_value) * num);
            }

            for (n = 0; n < num; ++n) {
                if (!xcoder_Read32(stream, &element->attrs[i].values[n].offset) ||
                    !xcoder_Read32(stream, &element->attrs[i].values[n].length))
                    return 0; /* XXX leak */
                    
            }
        }
    }

    if (!xcoder_Read32(stream, &element->nkids))
        return 0;

    if (FLAG_READ(stream) && element->nkids > 0) {
        element->kids = malloc(sizeof(HtmlElement) * element->nkids);
    }

    for (i = 0; i < element->nkids; ++i) {
        /* recursivly read the kids */
        xcoder_HtmlElement(stream, &element->kids[i]);
    }

    return 1;
}

#endif