/* -*- Mode: C++; tab-width: 20; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Anya client/server code
 *
 * The Initial Developer of the Original Code is Netscape Communications
 * Corporation. Portions created by the Initial Developer are
 * Copyright (C) 2003 the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *    Stuart Parmenter <pavlov@netscape.com>
 *    Brendan Eich <brendan@mozilla.org>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 */

/*
 * Buffered read/write code
 */

#include "xcoder.h"
#include "ozassert.h"


#ifdef XCODER_NO_WRITEFILE
#undef XCODER_USE_WRITEFILE
#else
#define XCODER_USE_WRITEFILE
#endif

#ifdef ANYA_NO_COMPRESSED_STREAM_READ
#undef ANYA_COMPRESSED_STREAM_READ
#else
#define ANYA_COMPRESSED_STREAM_READ
#endif

#ifdef ANYA_NO_COMPRESSED_STREAM_WRITE
#undef ANYA_COMPRESSED_STREAM_WRITE
#else
#define ANYA_COMPRESSED_STREAM_WRITE
#endif


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#if defined(_WIN32_WCE)
#include <winsock.h>
#undef XCODER_USE_WRITEFILE
#elif defined(_WIN32)
#include <winsock2.h>
#elif defined(OZ_PALMOS)
extern UInt16 netLibRefNum; /* from main.c */
#define htons(x) NetHToNS(x)
#define htonl(x) NetHToNL(x)
#define ntohs(x) NetNToHL(x)
#define ntohl(x) NetNToHS(x)
#endif


#ifdef XCODER_USE_WRITEFILE
#include <io.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#endif

int
xcoder_Open(XCoderStream *stream, int fd, int flags)
{
    stream->fd = fd;
    stream->flags = flags;
    stream->count = stream->cursor = 0;

#ifdef ANYA_COMPRESSED_STREAM
    stream->zstream = NULL;

    if (FLAG_COMPRESS(stream)) {
        stream->zstream = malloc(sizeof(z_stream));
        if (!stream->zstream)
            return 0;

        memset(stream->zstream, 0, sizeof(z_stream));

#ifdef ANYA_COMPRESSED_STREAM_READ

        if (FLAG_READ(stream)) {
            if (inflateInit2(stream->zstream,
                             MAX_WBITS) != Z_OK) {
                free(stream->zstream);
                return 0;
            }
        }
#endif
#ifdef ANYA_COMPRESSED_STREAM_WRITE
#ifdef ANYA_COMPRESSED_STREAM_READ
        else
#endif /* ANYA_COMPRESSED_STREAM_READ */
        if (FLAG_WRITE(stream)) {
            /* (strm, level, method, windowBits, memLevel, strategy) */
            if (deflateInit2(stream->zstream,
                             9,
                             Z_DEFLATED,
                             MAX_WBITS, 
                             8,
                             Z_DEFAULT_STRATEGY) != Z_OK) {
                free(stream->zstream);
                return 0;
            }
        }
#endif /* ANYA_COMPRESSED_STREAM_WRITE */
    }
#endif /* ANYA_COMPRESSED_STREAM */

    return 1;
}

void
xcoder_Close(XCoderStream *stream)
{
    OZ_ASSERT(stream->count == stream->cursor); /* this should probably just be a warning */

#ifdef ANYA_COMPRESSED_STREAM
    if (FLAG_COMPRESS(stream)) {
#ifdef ANYA_COMPRESSED_STREAM_READ
        if (FLAG_READ(stream))
            inflateEnd(stream->zstream);
#endif
#ifdef ANYA_COMPRESSED_STREAM_WRITE
#ifdef ANYA_COMPRESSED_STREAM_READ
        else
#endif /* ANYA_COMPRESSED_STREAM_READ */
        if (FLAG_WRITE(stream))
            deflateEnd(stream->zstream);
#endif /* ANYA_COMPRESSED_STREAM_WRITE */
        free(stream->zstream);
    }
#endif
}

#ifdef ANYA_COMPRESSED_STREAM_READ
static int
decompress_stream(XCoderStream *stream)
{
    int code;

    stream->zstream->next_in = stream->cbuffer;

    stream->zstream->next_out  = stream->buffer + stream->count;
    stream->zstream->avail_out = XCODER_BUFFER_SIZE - stream->count;

    code = inflate(stream->zstream, Z_SYNC_FLUSH);
    OZ_ASSERT(code >= 0);

    stream->count += (XCODER_BUFFER_SIZE - stream->count) - stream->zstream->avail_out;

    if (code == Z_STREAM_END)
        return 2;

    if (stream->zstream->avail_in > 0)
        memmove(stream->cbuffer, stream->zstream->next_in, stream->zstream->avail_in);

    return (code == Z_OK);
}
#endif

int
xcoder_FillBuffer(XCoderStream *stream)
{
    int resid, nbytes;

    OZ_ASSERT(FLAG_READ(stream));

    resid = stream->count - stream->cursor;
    OZ_ASSERT(resid >= 0);

    if (resid != 0)
        memmove(stream->buffer, stream->buffer + stream->cursor, resid);
    stream->count = resid;
    stream->cursor = 0;

    OZ_ASSERT(resid < XCODER_BUFFER_SIZE);
    if (resid >= XCODER_BUFFER_SIZE)
        return 0;

#ifdef ANYA_COMPRESSED_STREAM_READ
    if (FLAG_COMPRESS(stream)) {
        if (stream->zstream->avail_in > 0)
            if (decompress_stream(stream) == 2)
                return 1;

        while (stream->count < XCODER_BUFFER_SIZE) {
            nbytes = recv(stream->fd,
                          stream->cbuffer + stream->zstream->avail_in, 
                          XCODER_BUFFER_SIZE - stream->zstream->avail_in,
                          0);

            if (nbytes < 0)
                return 0;

            stream->zstream->avail_in += nbytes;
            if (decompress_stream(stream) == 2)
                return 1;
        }
        return stream->count != 0;
    } else {
#endif
#ifdef OZ_PALMOS
        {
        Err err;
        nbytes = NetLibReceive(netLibRefNum, stream->fd, 
                               stream->buffer + resid, 
                               XCODER_BUFFER_SIZE - resid,
                               0, NULL, NULL, -1, &err);
        }
#else
        nbytes = recv(stream->fd,
                      stream->buffer + resid, 
                      XCODER_BUFFER_SIZE - resid,
                      0);
#endif
        if (nbytes < 0)
            return 0;
        stream->count += nbytes;
        return stream->count != 0;

#ifdef ANYA_COMPRESSED_STREAM_READ
    }
#endif
}

static int
flush_send(XCoderStream *stream)
{
    int n = 0;
    while (stream->count < stream->cursor)
    {
#ifdef XCODER_USE_WRITEFILE
        if (FLAG_WRITE(stream)) {
#endif
#ifdef OZ_PALMOS
            Err err;
            n = NetLibSend(netLibRefNum, stream->fd, 
                           stream->buffer + stream->count,
                           stream->cursor - stream->count,
                           0, NULL, 0, -1, &err);
#else
            n = send(stream->fd,
                     stream->buffer + stream->count,
                     stream->cursor - stream->count,
                     0);
#endif
#ifdef XCODER_USE_WRITEFILE
        } else if (FLAG_WRITEFILE(stream)) {
            n = _write(stream->fd,
                       stream->buffer + stream->count,
                       stream->cursor - stream->count);
        }
#endif
        if (n == -1) { break; }
        stream->count += n;
    }
    stream->cursor = stream->count = 0;

    return n == -1 ? 0 : 1;
}

#ifdef ANYA_COMPRESSED_STREAM_WRITE
static int
compress_stream(XCoderStream *stream, int flush)
{
    int code;

    memcpy(stream->cbuffer, stream->buffer, stream->cursor);

    stream->zstream->next_in  = stream->cbuffer;
    stream->zstream->avail_in = stream->cursor;

    stream->zstream->next_out  = stream->buffer;
    stream->zstream->avail_out = XCODER_BUFFER_SIZE;

    code = deflate(stream->zstream, flush);

    OZ_ASSERT(code >= 0);
    /* if avail_out is 0, we need to call deflate again.
       we handle this by returning 0 here and dealing with it
       below */
    /* OZ_ASSERT(stream->zstream->avail_out != 0); */
    OZ_ASSERT(stream->zstream->avail_in == 0);

    stream->cursor = XCODER_BUFFER_SIZE - stream->zstream->avail_out;

    return (stream->zstream->avail_out == 0) ? 0 : 1;
}
#endif

int
xcoder_FlushBuffer(XCoderStream *stream)
{
    if (stream->cursor > 0) {
#ifdef ANYA_COMPRESSED_STREAM_WRITE
        if (FLAG_COMPRESS(stream)) {
            while (compress_stream(stream, Z_SYNC_FLUSH) == 0) {
                flush_send(stream);
            }

        }
#endif
        return flush_send(stream);
    }
    return 1;
}

/* XXX should this be moved to xcoder_Close ? */
int
xcoder_FlushBufferEnd(XCoderStream *stream)
{
    uint8 eof = 0;
    xcoder_Read8(stream, &eof);
#ifdef ANYA_COMPRESSED_STREAM_WRITE
    if (FLAG_COMPRESS(stream))
        compress_stream(stream, Z_FINISH);
#endif
    return flush_send(stream);
}

int
xcoder_CanAccess(XCoderStream *stream, int nbytes)
{
    int resid;
    if (FLAG_READ(stream)) {
        resid = stream->count - stream->cursor;

        while (resid < nbytes) {
            if (!xcoder_FillBuffer(stream))
                return 0;
            resid = stream->count - stream->cursor;
        }
    } else {
        resid = XCODER_BUFFER_SIZE - stream->cursor;
        if (resid < nbytes) {
            if (!xcoder_FlushBuffer(stream))
                return 0;
            resid = XCODER_BUFFER_SIZE - stream->cursor;
        }
    }
    return resid >= nbytes;
}

int
xcoder_Read8(XCoderStream *stream, uint8 *ip)
{
    if (!xcoder_CanAccess(stream, sizeof(*ip)))
        return 0;

    if (FLAG_READ(stream))
        *ip = stream->buffer[stream->cursor++];
    else {
        stream->buffer[stream->cursor] = *ip;
        stream->cursor++;
    }
    return 1;
}

int
xcoder_Read16(XCoderStream *stream, uint16 *ip)
{
    if (!xcoder_CanAccess(stream, sizeof(*ip)))
        return 0;

    if (FLAG_READ(stream)) {
        memcpy(ip, (stream->buffer + stream->cursor), sizeof(*ip));
        *ip = ntohs(*ip);
        stream->cursor += sizeof(*ip);
    } else {
        uint16 i = htons(*ip);
        memcpy(stream->buffer + stream->cursor, &i, sizeof(i));
        stream->cursor += sizeof(i);
    }
    return 1;
}

int
xcoder_Read32(XCoderStream *stream, uint32 *ip)
{
    if (!xcoder_CanAccess(stream, sizeof(*ip)))
        return 0;

    if (FLAG_READ(stream)) {
        memcpy(ip, (stream->buffer + stream->cursor), sizeof(*ip));
        *ip = ntohl(*ip);
        stream->cursor += sizeof(*ip);
    } else {
        uint32 i = htonl(*ip);
        memcpy(stream->buffer + stream->cursor, &i, sizeof(i));
        stream->cursor += sizeof(i);
    }
    return 1;
}

int
xcoder_ReadBytes(XCoderStream *stream, uint32 *len, uint8 **sp)
{
    uint8 *cp, *end;
    int resid;

    if (!xcoder_Read32(stream, len))
        return 0;

    if (FLAG_READ(stream)) {

        if (*len == 0) {
            *sp = NULL;
            return 1;
        }

        /* XXX if malloc fails, we should probably continue to read
           in to a small buffer until we've read all of len. */
        *sp = (uint8 *)malloc(*len + 1);
        if (!*sp)
            return 0;

        cp = *sp;
        end = cp + *len;
        while (cp < end) {
            resid = stream->count - stream->cursor;
            if (cp + resid > end)
                resid = end - cp;

            memcpy(cp, (stream->buffer + stream->cursor), resid);
            cp += resid;
            stream->cursor += resid;

            if (cp == end)
                break;

            if (stream->cursor == stream->count && !xcoder_FillBuffer(stream))
                return 0; /* XXX leak */
        }
        *cp = '\0';

    } else {
        if (*len == 0)
            return 1;

        cp = *sp;
        end = cp + *len;
        while (cp < end) {
            resid = XCODER_BUFFER_SIZE - stream->cursor;
            if (cp + resid > end)
                resid = end - cp;

            memcpy(stream->buffer + stream->cursor, cp, resid);
            cp += resid;
            stream->cursor += resid;

            if (cp == end)
                break;

            if (XCODER_BUFFER_SIZE == stream->cursor && !xcoder_FlushBuffer(stream))
                return 0;
        }

    }
    return 1;
}

int
xcoder_ReadString(XCoderStream *stream, char **sp)
{
    uint32 len = FLAG_READ(stream) ? 0 : (*sp ? strlen(*sp) : 0);
    return xcoder_ReadNString(stream, &len, sp);
}

int
xcoder_ReadNString(XCoderStream *stream, uint32 *len, char **sp)
{
    return xcoder_ReadBytes(stream, len, (uint8**)sp);
}

