/* -*- Mode: C++; tab-width: 20; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Anya server code.
 *
 * The Initial Developer of the Original Code is Netscape Communications
 * Corporation. Portions created by the Initial Developer are
 * Copyright (C) 2003 the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *    Stuart Parmenter <pavlov@netscape.com>
 *    Joe Hewitt <hewitt@netscape.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 */

#include "anyaRenderingContext.h"

#include "nsRect.h"
#include "nsString.h"
#include "nsTransform2D.h"
#include "nsIRegion.h"
#include "nsIServiceManager.h"
#include "nsIInterfaceRequestorUtils.h"
#include "nsGfxCIID.h"

#include "imgIContainer.h"
#include "gfxIImageFrame.h"
#include "nsIImage.h"

#include "anyaFontMetrics.h"

#define IMAGE_DOWNSAMPLE

#ifdef IMAGE_DOWNSAMPLE
#include "anyaImageDownSample.h"
#endif

#include <io.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "xcoder.h"
#include "prims.h"

static NS_DEFINE_CID(kRegionCID, NS_REGION_CID);

//////////////////////////////////////////////////////////////////////
//// anyaGraphicsState

class anyaGraphicsState
{
public:
    anyaGraphicsState();
    anyaGraphicsState(anyaGraphicsState &aState);
    ~anyaGraphicsState();

    anyaGraphicsState *mNext;
    nsTransform2D *mMatrix;
    nsCOMPtr<nsIRegion> mClipRegion;
    nsCOMPtr<nsIFontMetrics> mFontMetrics;
    nscolor mColor;
    nsLineStyle mLineStyle;
};

anyaGraphicsState :: anyaGraphicsState()
{
    mNext = nsnull;
    mMatrix = nsnull;
    mColor = NS_RGB(0, 0, 0);
    mLineStyle = nsLineStyle_kSolid;
    mFontMetrics = nsnull;
}

anyaGraphicsState :: ~anyaGraphicsState()
{
}

//////////////////////////////////////////////////////////////////////

NS_IMPL_ISUPPORTS1(anyaRenderingContext, nsIRenderingContext)

anyaRenderingContext::anyaRenderingContext() :
    mLineStyle(nsLineStyle_kNone),
    mColor(0),
    mState(nsnull),
    mSurfaceSize(0, 0)
{
    mTranMatrix = new nsTransform2D();

    mADTM = do_GetService(NS_ANYADATAXFERMGR_CONTRACTID);

    PushState();

    mP2T = 1.0f;
}

anyaRenderingContext::~anyaRenderingContext()
{
    PRInt32 cnt = mStateCache.Count();

    while (--cnt >= 0)
    {
        PRBool  clipstate;
        PopState(clipstate);
    }

    delete mTranMatrix;

    DownSampleCleanup();
}

//////////////////////////////////////////////////////////////////////
//// nsIRenderingContext

NS_IMETHODIMP
anyaRenderingContext::Init(nsIDeviceContext* aContext, nsIWidget *aWidget)
{
    mDeviceContext = aContext;
    mWidget = aWidget;

    float app2dev;
    mDeviceContext->GetAppUnitsToDevUnits(app2dev);
    mTranMatrix->AddScale(app2dev, app2dev);
    mDeviceContext->GetDevUnitsToAppUnits(mP2T);

    if (aWidget) {
        nsRect bounds;
        aWidget->GetBounds(bounds);
        mSurfaceSize.width = bounds.width;
        mSurfaceSize.height = bounds.height;
    }
    return NS_OK;
}

NS_IMETHODIMP
anyaRenderingContext::Init(nsIDeviceContext* aContext, nsDrawingSurface aSurface)
{
    mDeviceContext = aContext;
    return NS_OK;
}

NS_IMETHODIMP
anyaRenderingContext::Reset(void)
{
    return NS_OK;
}

NS_IMETHODIMP
anyaRenderingContext::GetDeviceContext(nsIDeviceContext *& aDeviceContext)
{
    aDeviceContext = mDeviceContext;
    NS_IF_ADDREF(aDeviceContext);
    return NS_OK;
}

NS_IMETHODIMP
anyaRenderingContext::LockDrawingSurface(PRInt32 aX, PRInt32 aY,
                                         PRUint32 aWidth, PRUint32 aHeight,
                                         void **aBits, PRInt32 *aStride,
                                         PRInt32 *aWidthBytes,
                                         PRUint32 aFlags)
{
    return NS_OK;
}

NS_IMETHODIMP
anyaRenderingContext::UnlockDrawingSurface(void)
{

    return NS_OK;
}

NS_IMETHODIMP
anyaRenderingContext::SelectOffScreenDrawingSurface(nsDrawingSurface aSurface)
{

    return NS_OK;
}

NS_IMETHODIMP
anyaRenderingContext::GetDrawingSurface(nsDrawingSurface *aSurface)
{

    return NS_OK;
}

NS_IMETHODIMP
anyaRenderingContext::GetHints(PRUint32& aResult)
{
    aResult = 0;

    return NS_OK;
}

NS_IMETHODIMP
anyaRenderingContext::PushState(void)
{
    anyaGraphicsState *state = new anyaGraphicsState;

    if (!state)
        return NS_ERROR_FAILURE;

    state->mMatrix = mTranMatrix;

    if (nsnull == mTranMatrix)
        mTranMatrix = new nsTransform2D();
    else
        mTranMatrix = new nsTransform2D(mTranMatrix);

    state->mClipRegion = mClipRegion;

    state->mFontMetrics = mFontMetrics;

    state->mColor = mColor;
    state->mLineStyle = mLineStyle;

    mStateCache.AppendElement(state);

    return NS_OK;
}

NS_IMETHODIMP
anyaRenderingContext::PopState(PRBool &aClipEmpty)
{
    PRUint32 cnt = mStateCache.Count();
    anyaGraphicsState *state;

    if (cnt > 0) {
        state = (anyaGraphicsState *)mStateCache.ElementAt(cnt - 1);
        mStateCache.RemoveElementAt(cnt - 1);

        if (state->mMatrix) {
            if (mTranMatrix)
                delete mTranMatrix;
            mTranMatrix = state->mMatrix;
        }

        mClipRegion = state->mClipRegion;

        if (state->mFontMetrics && (mFontMetrics != state->mFontMetrics))
            SetFont(state->mFontMetrics);

        if (state->mColor != mColor)
            SetColor(state->mColor);    

        if (state->mLineStyle != mLineStyle)
            SetLineStyle(state->mLineStyle);

        delete state;
    }

    if (mClipRegion)
        aClipEmpty = mClipRegion->IsEmpty();
    else
        aClipEmpty = PR_TRUE;

    return NS_OK;
}

NS_IMETHODIMP
anyaRenderingContext::IsVisibleRect(const nsRect& aRect, PRBool &aIsVisible)
{
    aIsVisible = PR_TRUE;
    return NS_OK;
}

NS_IMETHODIMP
anyaRenderingContext::SetClipRect(const nsRect& aRect, nsClipCombine aCombine,
                                  PRBool &aClipEmpty)
{
    nsCOMPtr<nsIRegion> region(do_CreateInstance(kRegionCID));

    nsRect r = aRect;
    mTranMatrix->TransformCoord(&r.x, &r.y, &r.width, &r.height);

    region->SetTo(r.x, r.y, r.width, r.height);
    return SetClipRegion(*(region.get()), aCombine, aClipEmpty);
}

NS_IMETHODIMP
anyaRenderingContext::GetClipRect(nsRect &aRect, PRBool &aHasLocalClip)
{
    PRInt32 x, y, w, h;
  
    if (!mClipRegion)
        return NS_ERROR_FAILURE;

    if (!mClipRegion->IsEmpty()) {
        mClipRegion->GetBoundingBox(&x,&y,&w,&h);
        aRect.SetRect(x,y,w,h);
        aHasLocalClip = PR_TRUE;
    } else {
        aRect.SetRect(0,0,0,0);
        aHasLocalClip = PR_FALSE;
    }

    return NS_OK;
}

NS_IMETHODIMP
anyaRenderingContext::SetLineStyle(nsLineStyle aLineStyle)
{
    mLineStyle = aLineStyle;
    return NS_OK;
}

NS_IMETHODIMP
anyaRenderingContext::GetLineStyle(nsLineStyle &aLineStyle)
{
    aLineStyle = mLineStyle;
    return NS_OK;
}

NS_IMETHODIMP
anyaRenderingContext::GetPenMode(nsPenMode &aPenMode)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP
anyaRenderingContext::SetPenMode(nsPenMode aPenMode)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP
anyaRenderingContext::SetClipRegion(const nsIRegion& aRegion,
                                    nsClipCombine aCombine, PRBool &aClipEmpty)
{
    PRUint32 cnt = mStateCache.Count();
    anyaGraphicsState *state = nsnull;

    if (cnt > 0) {
        state = (anyaGraphicsState *)mStateCache.ElementAt(cnt - 1);
    }

    if (state && state->mClipRegion) {
        if (state->mClipRegion == mClipRegion) {
            nsCOMPtr<nsIRegion> tmpRgn;
            GetClipRegion(getter_AddRefs(tmpRgn));
            mClipRegion = tmpRgn;
        }
    }

    if (!mClipRegion) {
        mClipRegion = do_CreateInstance(kRegionCID);
        if (mClipRegion) {
            mClipRegion->Init();
            mClipRegion->SetTo(0,0,mSurfaceSize.width,mSurfaceSize.width);
        }
    }

    switch(aCombine)
    {
    case nsClipCombine_kIntersect:
        mClipRegion->Intersect(aRegion);
        break;
    case nsClipCombine_kUnion:
        mClipRegion->Union(aRegion);
        break;
    case nsClipCombine_kSubtract:
        mClipRegion->Subtract(aRegion);
        break;
    case nsClipCombine_kReplace:
        mClipRegion->SetTo(aRegion);
        break;
    }

    aClipEmpty = mClipRegion->IsEmpty();

    return NS_OK;
}

NS_IMETHODIMP
anyaRenderingContext::CopyClipRegion(nsIRegion &aRegion)
{
    if (!mClipRegion)
        return NS_ERROR_FAILURE;

    aRegion.SetTo(*mClipRegion);
    return NS_OK;
}

NS_IMETHODIMP
anyaRenderingContext::GetClipRegion(nsIRegion **aRegion)
{
    nsresult rv = NS_ERROR_FAILURE;

    if (!aRegion || !mClipRegion)
        return NS_ERROR_NULL_POINTER;

    if (mClipRegion) {
        if (*aRegion) { // copy it, they should be using CopyClipRegion 
            (*aRegion)->SetTo(*mClipRegion);
            rv = NS_OK;
        } else {
            nsCOMPtr<nsIRegion> newRegion = do_CreateInstance(kRegionCID, &rv);
            if (NS_SUCCEEDED(rv)) {
                newRegion->Init();
                newRegion->SetTo(*mClipRegion);
                NS_ADDREF(*aRegion = newRegion);
            }
        }
    } else {
        rv = NS_ERROR_FAILURE;
    }

    return rv;
}

NS_IMETHODIMP
anyaRenderingContext::SetColor(nscolor aColor)
{
    if (aColor == mColor)
        return NS_OK;

#ifdef DEBUG
    printf("SET COLOR %d, %d, %d\n", NS_GET_R(aColor), NS_GET_G(aColor), NS_GET_B(aColor));
#endif

    mColor = aColor;

    struct Color c;
    c.r = NS_GET_R(aColor);
    c.g = NS_GET_G(aColor);
    c.b = NS_GET_B(aColor);
    xcoder_Color(mADTM->GetXCoder(), &c);
    mADTM->Inc(sizeof(struct Color));

    return NS_OK;
}

NS_IMETHODIMP
anyaRenderingContext::GetColor(nscolor &aColor) const
{
    aColor = mColor;

    return NS_OK;
}

NS_IMETHODIMP
anyaRenderingContext::SetFont(const nsFont& aFont, nsIAtom* aLangGroup)
{
    mDeviceContext->GetMetricsFor(aFont, aLangGroup, *getter_AddRefs(mFontMetrics));
    return NS_OK;
}

NS_IMETHODIMP
anyaRenderingContext::SetFont(nsIFontMetrics *aFontMetrics)
{
    mFontMetrics = aFontMetrics;
    return NS_OK;
}

NS_IMETHODIMP
anyaRenderingContext::GetFontMetrics(nsIFontMetrics *&aFontMetrics)
{
    aFontMetrics = mFontMetrics;
    NS_IF_ADDREF(aFontMetrics);
    return NS_OK;
}

NS_IMETHODIMP
anyaRenderingContext::Translate(nscoord aX, nscoord aY)
{
    mTranMatrix->AddTranslation((float)aX,(float)aY);
    return NS_OK;
}

NS_IMETHODIMP
anyaRenderingContext::Scale(float aSx, float aSy)
{
    mTranMatrix->AddScale(aSx, aSy);
    return NS_OK;
}

NS_IMETHODIMP
anyaRenderingContext::GetCurrentTransform(nsTransform2D *&aTransform)
{
    aTransform = mTranMatrix;
    return NS_OK;
}

NS_IMETHODIMP
anyaRenderingContext::CreateDrawingSurface(const nsRect &aBounds,
                                           PRUint32 aSurfFlags,
                                           nsDrawingSurface &aSurface)
{

    return NS_OK;
}

NS_IMETHODIMP
anyaRenderingContext::DestroyDrawingSurface(nsDrawingSurface aDS)
{

    return NS_OK;
}

NS_IMETHODIMP
anyaRenderingContext::DrawLine(nscoord aX0, nscoord aY0, nscoord aX1, nscoord aY1)
{
    mTranMatrix->TransformCoord(&aX0,&aY0);
    mTranMatrix->TransformCoord(&aX1,&aY1);

#ifdef DEBUG
    printf("DRAW LINE %d, %d, %d, %d\n", aX0, aY0, aX1, aY1);
#endif

    struct Line l;
    l.x1 = aX0;
    l.y1 = aY0;
    l.x2 = aX1;
    l.y2 = aY1;
    xcoder_Line(mADTM->GetXCoder(), &l);
    mADTM->Inc(sizeof(struct Line));

    return NS_OK;
}

NS_IMETHODIMP
anyaRenderingContext::DrawStdLine(nscoord aX0, nscoord aY0, nscoord aX1, nscoord aY1)
{
    return NS_OK;
}

NS_IMETHODIMP
anyaRenderingContext::DrawPolyline(const nsPoint aPoints[],
                                   PRInt32 aNumPoints)
{
    ozPoint *pts = new ozPoint[aNumPoints];
    for (PRInt32 i = 0; i < aNumPoints; ++i) {
        int x = aPoints[i].x;
        int y = aPoints[i].y;
        mTranMatrix->TransformCoord(&x, &y);
        pts[i].x = x;
        pts[i].y = y;
    }
    struct Polyline p;
    p.npoints = aNumPoints;
    p.points = pts;
    xcoder_Polyline(mADTM->GetXCoder(), &p);
    mADTM->Inc(sizeof(struct Polyline));
    delete[] pts;

    return NS_OK;
}

NS_IMETHODIMP
anyaRenderingContext::DrawRect(const nsRect& aRect)
{
    return DrawRect(aRect.x, aRect.y, aRect.width, aRect.height);
}

NS_IMETHODIMP
anyaRenderingContext::DrawRect(nscoord aX, nscoord aY, nscoord aWidth, nscoord aHeight)
{
    mTranMatrix->TransformCoord(&aX, &aY, &aWidth, &aHeight);

#ifdef DEBUG
    printf("DRAW RECT %d, %d, %d, %d\n", aX, aY, aWidth, aHeight);
#endif

    nsRect rect;
    nsCOMPtr<nsIRegion> rgn;
    GetClipRegion(getter_AddRefs(rgn));
    rgn->Intersect(aX, aY, aWidth, aHeight);
    rgn->GetBoundingBox(&aX, &aY, &aWidth, &aHeight);

    struct Rect r;
    r.filled = FALSE;
    r.x = aX;
    r.y = aY;
    r.w = aWidth;
    r.h = aHeight;
    xcoder_Rect(mADTM->GetXCoder(), &r);
    mADTM->Inc(sizeof(struct Rect));

    return NS_OK;
}

NS_IMETHODIMP
anyaRenderingContext::FillRect(const nsRect& aRect)
{
    return FillRect(aRect.x, aRect.y, aRect.width, aRect.height);
}

NS_IMETHODIMP
anyaRenderingContext::FillRect(nscoord aX, nscoord aY, nscoord aWidth, nscoord aHeight)
{
    mTranMatrix->TransformCoord(&aX, &aY, &aWidth, &aHeight);

    /* XXX call DrawLine when we have aHeight of 1 */

#ifdef DEBUG
    printf("FILL RECT %d, %d, %d, %d\n", aX, aY, aWidth, aHeight);
#endif

    nsRect rect;
    nsCOMPtr<nsIRegion> rgn;
    GetClipRegion(getter_AddRefs(rgn));
    if (rgn) {
        rgn->Intersect(aX, aY, aWidth, aHeight);
        rgn->GetBoundingBox(&aX, &aY, &aWidth, &aHeight);
    }

    struct Rect r;
    r.filled = TRUE;
    r.x = aX;
    r.y = aY;
    r.w = aWidth;
    r.h = aHeight;
    xcoder_Rect(mADTM->GetXCoder(), &r);
    mADTM->Inc(sizeof(struct Rect));

    return NS_OK;
}

NS_IMETHODIMP
anyaRenderingContext::InvertRect(const nsRect& aRect)
{
    return InvertRect(aRect.x, aRect.y, aRect.width, aRect.height);
}

NS_IMETHODIMP
anyaRenderingContext::InvertRect(nscoord aX, nscoord aY, nscoord aWidth, nscoord aHeight)
{
    return NS_OK;
}

NS_IMETHODIMP
anyaRenderingContext::FlushRect(const nsRect& aRect)
{
    return FlushRect(aRect.x, aRect.y, aRect.width, aRect.height);
}

NS_IMETHODIMP
anyaRenderingContext::FlushRect(nscoord aX, nscoord aY, nscoord aWidth, nscoord aHeight)
{
    return NS_OK;
}
  
NS_IMETHODIMP
anyaRenderingContext::DrawPolygon(const nsPoint aPoints[], PRInt32 aNumPoints)
{
    /* XXX check and make sure that pts[0] == aPoints[aNumPoints] */
    ozPoint *pts = new ozPoint[aNumPoints];
    for (PRInt32 i = 0; i < aNumPoints; ++i) {
        int x = aPoints[i].x;
        int y = aPoints[i].y;
        mTranMatrix->TransformCoord(&x, &y);
        pts[i].x = x;
        pts[i].y = y;
    }
    struct Polygon p;
    p.filled = FALSE;
    p.npoints = aNumPoints;
    p.points = pts;
    xcoder_Polygon(mADTM->GetXCoder(), &p);
    mADTM->Inc(sizeof(struct Polygon));
    delete[] pts;

    return NS_OK;
}

NS_IMETHODIMP
anyaRenderingContext::FillPolygon(const nsPoint aPoints[], PRInt32 aNumPoints)
{
    PRInt32 i;
    ozPoint *pts = new ozPoint[aNumPoints + 1];
    for (i = 0; i < aNumPoints; ++i) {
        int x = aPoints[i].x;
        int y = aPoints[i].y;
        mTranMatrix->TransformCoord(&x, &y);
        pts[i].x = x;
        pts[i].y = y;
    }
    /* we must send the first point == last point */
    pts[i].x = pts[0].x;
    pts[i].y = pts[0].y;

    struct Polygon p;
    p.filled = TRUE;
    p.npoints = aNumPoints + 1;
    p.points = pts;
    xcoder_Polygon(mADTM->GetXCoder(), &p);
    mADTM->Inc(sizeof(struct Polygon));
    delete[] pts;

    return NS_OK;
}

NS_IMETHODIMP
anyaRenderingContext::RasterPolygon(const nsPoint aPoints[], PRInt32 aNumPoints)
{
    return NS_OK;
}

NS_IMETHODIMP
anyaRenderingContext::FillStdPolygon(const nsPoint aPoints[], PRInt32 aNumPoints)
{
    return NS_OK;
}

NS_IMETHODIMP
anyaRenderingContext::DrawEllipse(const nsRect& aRect)
{
    return DrawEllipse(aRect.x, aRect.y, aRect.width, aRect.height);
}

NS_IMETHODIMP
anyaRenderingContext::DrawEllipse(nscoord aX, nscoord aY, nscoord aWidth, nscoord aHeight)
{
    mTranMatrix->TransformCoord(&aX, &aY, &aWidth, &aHeight);

    return NS_OK;
}

NS_IMETHODIMP
anyaRenderingContext::FillEllipse(const nsRect& aRect)
{
    return FillEllipse(aRect.x, aRect.y, aRect.width, aRect.height);
}

NS_IMETHODIMP
anyaRenderingContext::FillEllipse(nscoord aX, nscoord aY, nscoord aWidth, nscoord aHeight)
{
    mTranMatrix->TransformCoord(&aX, &aY, &aWidth, &aHeight);

    return NS_OK;
}

NS_IMETHODIMP
anyaRenderingContext::DrawArc(const nsRect& aRect,
                                            float aStartAngle, float aEndAngle)
{
    return DrawArc(aRect.x, aRect.y, aRect.width, aRect.height, aStartAngle, aEndAngle);
}

NS_IMETHODIMP
anyaRenderingContext::DrawArc(nscoord aX, nscoord aY, nscoord aWidth, nscoord aHeight,
                              float aStartAngle, float aEndAngle)
{
    mTranMatrix->TransformCoord(&aX, &aY, &aWidth, &aHeight);

    return NS_OK;
}

NS_IMETHODIMP
anyaRenderingContext::FillArc(const nsRect& aRect,
                              float aStartAngle, float aEndAngle)
{
    return FillArc(aRect.x, aRect.y, aRect.width, aRect.height, aStartAngle, aEndAngle);
}

NS_IMETHODIMP
anyaRenderingContext::FillArc(nscoord aX, nscoord aY, nscoord aWidth, nscoord aHeight,
                              float aStartAngle, float aEndAngle)
{
    mTranMatrix->TransformCoord(&aX, &aY, &aWidth, &aHeight);

    return NS_OK;
}

NS_IMETHODIMP
anyaRenderingContext::GetWidth(char aC, nscoord &aWidth)
{
    if (aC == ' ' && mFontMetrics)
        return mFontMetrics->GetSpaceWidth(aWidth);

    return GetWidth(&aC, 1, aWidth);
}

NS_IMETHODIMP
anyaRenderingContext::GetWidth(PRUnichar aC, nscoord &aWidth,
                               PRInt32 *aFontID)
{
    return GetWidth(&aC, 1, aWidth, aFontID);
}

NS_IMETHODIMP
anyaRenderingContext::GetWidth(const nsString& aString, nscoord &aWidth,
                               PRInt32 *aFontID)
{
    return GetWidth(aString.get(), aString.Length(), aWidth, aFontID);
}

NS_IMETHODIMP
anyaRenderingContext::GetWidth(const char* aString, nscoord& aWidth)
{
    return GetWidth(aString, strlen(aString), aWidth);
}

NS_IMETHODIMP
anyaRenderingContext::GetWidth(const char* aString, PRUint32 aLength,
                               nscoord& aWidth)
{
    if (aLength == 0) {
        aWidth = 0;
        return NS_OK;
    }

    aWidth = NS_REINTERPRET_CAST(anyaFontMetrics*, mFontMetrics.get())->MeasureString(aString, aLength);
    return NS_OK;
}

NS_IMETHODIMP
anyaRenderingContext::GetWidth(const PRUnichar *aString, PRUint32 aLength,
                               nscoord &aWidth, PRInt32 *aFontID)
{
    if (aLength == 0) {
        aWidth = 0;
        return NS_OK;
    }

    aWidth = NS_REINTERPRET_CAST(anyaFontMetrics*, mFontMetrics.get())->MeasureString(aString, aLength);
    return NS_OK;
}

NS_IMETHODIMP
anyaRenderingContext::GetTextDimensions(const char* aString, PRUint32 aLength,
                                        nsTextDimensions& aDimensions)
{
  mFontMetrics->GetMaxAscent(aDimensions.ascent);
  mFontMetrics->GetMaxDescent(aDimensions.descent);
  return GetWidth(aString, aLength, aDimensions.width);
}

NS_IMETHODIMP
anyaRenderingContext::GetTextDimensions(const PRUnichar* aString,
                                        PRUint32 aLength,
                                        nsTextDimensions& aDimensions,
                                        PRInt32* aFontID)
{
  mFontMetrics->GetMaxAscent(aDimensions.ascent);
  mFontMetrics->GetMaxDescent(aDimensions.descent);
  return GetWidth(aString, aLength, aDimensions.width, aFontID);
}

NS_IMETHODIMP
anyaRenderingContext::GetTextDimensions(const char*       aString,
                                        PRInt32           aLength,
                                        PRInt32           aAvailWidth,
                                        PRInt32*          aBreaks,
                                        PRInt32           aNumBreaks,
                                        nsTextDimensions& aDimensions,
                                        PRInt32&          aNumCharsFit,
                                        nsTextDimensions& aLastWordDimensions,
                                        PRInt32*          aFontID)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP
anyaRenderingContext::GetTextDimensions(const PRUnichar*  aString,
                                        PRInt32           aLength,
                                        PRInt32           aAvailWidth,
                                        PRInt32*          aBreaks,
                                        PRInt32           aNumBreaks,
                                        nsTextDimensions& aDimensions,
                                        PRInt32&          aNumCharsFit,
                                        nsTextDimensions& aLastWordDimensions,
                                        PRInt32*          aFontID)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}


NS_IMETHODIMP
anyaRenderingContext::DrawString(const char *aString, PRUint32 aLength,
                                 nscoord aX, nscoord aY,
                                 const nscoord* aSpacing)
{
    mTranMatrix->TransformCoord(&aX, &aY);

#ifdef DEBUG
    printf("DRAW STRING %d,%d\n", aX, aY);
#endif
    PRUint32 id = idmap_AddString(mADTM->GetIdMap(), (char *)aString, aLength);

    struct Text t;
    t.x = aX;
    t.y = aY - 11;
    t.id = id;
    xcoder_Text(mADTM->GetXCoder(), &t);
    mADTM->Inc(sizeof(struct Text));

    return NS_OK;
}

NS_IMETHODIMP
anyaRenderingContext::DrawString(const PRUnichar *aString, PRUint32 aLength,
                                 nscoord aX, nscoord aY,
                                 PRInt32 aFontID,
                                 const nscoord* aSpacing)
{
    mTranMatrix->TransformCoord(&aX, &aY);

#ifdef DEBUG
    printf("DRAW STRING %d,%d\n", aX, aY);
#endif
    nsCAutoString s = NS_ConvertUCS2toUTF8(aString, aLength);
    PRUint32 id = idmap_AddString(mADTM->GetIdMap(), (char *)s.get(), s.Length());

    struct Text t;
    t.x = aX;
    t.y = aY - 11;
    t.id = id;
    xcoder_Text(mADTM->GetXCoder(), &t);
    mADTM->Inc(sizeof(struct Text));

    return NS_OK;
}

NS_IMETHODIMP
anyaRenderingContext::DrawString(const nsString& aString,
                                 nscoord aX, nscoord aY,
                                 PRInt32 aFontID,
                                 const nscoord* aSpacing)
{
    return DrawString(aString.get(), aString.Length(), aX, aY, aFontID, aSpacing);
}

NS_IMETHODIMP
anyaRenderingContext::DrawPath(nsPathPoint aPointArray[],PRInt32 aNumPts)
{
    return NS_OK;
}

NS_IMETHODIMP
anyaRenderingContext::FillPath(nsPathPoint aPointArray[],PRInt32 aNumPts)
{
    return NS_OK;
}

NS_IMETHODIMP
anyaRenderingContext::CopyOffScreenBits(nsDrawingSurface aSrcSurf,
                                        PRInt32 aSrcX, PRInt32 aSrcY,
                                        const nsRect &aDestBounds,
                                        PRUint32 aCopyFlags)
{

    return NS_OK;
}

NS_IMETHODIMP
anyaRenderingContext::RetrieveCurrentNativeGraphicData(PRUint32 * ngd)
{

    return NS_OK;
}

NS_IMETHODIMP
anyaRenderingContext::UseBackbuffer(PRBool* aUseBackbuffer)
{
    *aUseBackbuffer = PR_FALSE;
    return NS_OK;
}

NS_IMETHODIMP
anyaRenderingContext::GetBackbuffer(const nsRect &aRequestedSize,
                                    const nsRect &aMaxSize,
                                    nsDrawingSurface &aBackbuffer)
{
    return NS_OK;
}

NS_IMETHODIMP
anyaRenderingContext::ReleaseBackbuffer(void)
{
    return NS_OK;
}

NS_IMETHODIMP
anyaRenderingContext::DestroyCachedBackbuffer(void)
{
    return NS_OK;
}

#ifdef MOZ_MATHML
NS_IMETHODIMP 
anyaRenderingContext::GetBoundingMetrics(const char*        aString,
                                         PRUint32           aLength,
                                         nsBoundingMetrics& aBoundingMetrics)
{
  return NS_OK;
}

NS_IMETHODIMP
anyaRenderingContext::GetBoundingMetrics(const PRUnichar*   aString,
                                         PRUint32           aLength,
                                         nsBoundingMetrics& aBoundingMetrics,
                                         PRInt32*           aFontID)
{
    return NS_OK;
}
#endif // MOZ_MATHML

static void
AddAlphaData(nsIImage* aImage, struct Image* aXImage)
{
    if (aImage->GetHasAlphaMask()) {
        PRInt32 alphaLineStride = aImage->GetAlphaLineStride();
        PRInt32 alphaByteCount = alphaLineStride * aXImage->h;
        PRInt32 bufferSize = aXImage->nbytes + alphaByteCount;

        PRUint8* buffer = (PRUint8*)malloc(bufferSize);

        PRUint8* alphaPtr = buffer + aXImage->nbytes;
        memcpy(buffer, aXImage->bytes, aXImage->nbytes);

        PRUint8* alphaBits = aImage->GetAlphaBits();
        memcpy(alphaPtr, alphaBits, alphaByteCount);

        aXImage->bytes = buffer;
        aXImage->nbytes = bufferSize;
    }
}

NS_IMETHODIMP
anyaRenderingContext::DrawImage(imgIContainer *aImage, const nsRect * aSrcRect,
                                const nsPoint * aDestPoint)
{
    nsPoint pt;
    nsRect sr;

    float fX = float(aDestPoint->x), fY = float(aDestPoint->y);
    mTranMatrix->Transform(&fX, &fY);

    pt.x = NSToCoordRound(fX);
    pt.y = NSToCoordRound(fY);

    sr = *aSrcRect;
    mTranMatrix->TransformCoord(&sr.x, &sr.y, &sr.width, &sr.height);

    sr.x = aSrcRect->x;
    sr.y = aSrcRect->y;
    mTranMatrix->TransformNoXLateCoord(&sr.x, &sr.y);

    nsCOMPtr<gfxIImageFrame> iframe;
    aImage->GetCurrentFrame(getter_AddRefs(iframe));
    if (!iframe) return NS_ERROR_FAILURE;

    nsCOMPtr<nsIImage> img(do_GetInterface(iframe));
    if (!img) return NS_ERROR_FAILURE;

    // For Bug 87819
    // iframe may want image to start at different position, so adjust
    nsRect iframeRect;
    iframe->GetRect(iframeRect);

    if (iframeRect.y > 0) {
        sr.y -= iframeRect.y;
        if (sr.y < 0) {
            pt.y = pt.y - sr.y;
            sr.height = sr.height + sr.y;
            if (sr.height < 0)
                return NS_OK;
            sr.y = 0;
        } else if (sr.y > iframeRect.height) {
            return NS_OK;
        }
    }
  
    if (iframeRect.x > 0) {
        sr.x -= iframeRect.x;
        if (sr.x < 0) {
            pt.x = pt.x - sr.x;
            sr.width = sr.width + sr.x;
            if (sr.width < 0)
                return NS_OK;
            sr.x = 0;
        } else if (sr.x > iframeRect.width) {
            return NS_OK;
        }
    }


    struct Image ximg;
    memset(&ximg, 0, sizeof(struct Image));

    img->LockImagePixels(PR_FALSE);

    ximg.x = pt.x;
    ximg.y = pt.y;
    ximg.w = ximg.destw = img->GetWidth();
    ximg.h = ximg.desth = img->GetHeight();
    ximg.bytes = img->GetBits();
    ximg.nbytes = img->GetLineStride() * img->GetHeight(); // XXXben ??? check to see if DHeight is what we should be using here. 
    ximg.bpp = img->GetBytesPix();

    struct Request* req = mADTM->GetRequest();
    DownSampleImage(&ximg, req->device.display);

    AddAlphaData(img, &ximg);

    xcoder_Image(mADTM->GetXCoder(), &ximg);

    if (ximg.bytes != img->GetBits()) {
        free(ximg.bytes);
        ximg.bytes = NULL;
    }

    mADTM->Inc(sizeof(struct Image));
    img->UnlockImagePixels(PR_FALSE);

    return NS_OK;
}

NS_IMETHODIMP
anyaRenderingContext::DrawScaledImage(imgIContainer *aImage, const nsRect * aSrcRect,
                                      const nsRect * aDestRect)
{
    nsRect dr;
    nsRect sr;

    dr = *aDestRect;
    mTranMatrix->TransformCoord(&dr.x, &dr.y, &dr.width, &dr.height);

    sr = *aSrcRect;
    mTranMatrix->TransformCoord(&sr.x, &sr.y, &sr.width, &sr.height);

    if (sr.width <= 0 || sr.height <= 0 || dr.width <= 0 || dr.height <= 0)
    return NS_OK;

    sr.x = aSrcRect->x;
    sr.y = aSrcRect->y;
    mTranMatrix->TransformNoXLateCoord(&sr.x, &sr.y);

    nsCOMPtr<gfxIImageFrame> iframe;
    aImage->GetCurrentFrame(getter_AddRefs(iframe));
    if (!iframe) return NS_ERROR_FAILURE;

    nsCOMPtr<nsIImage> img(do_GetInterface(iframe));
    if (!img) return NS_ERROR_FAILURE;

    // For Bug 87819
    // iframe may want image to start at different position, so adjust
    nsRect iframeRect;
    iframe->GetRect(iframeRect);

    if (iframeRect.x > 0) {
        float xScaleRatio = float(dr.width) / sr.width;
        sr.x -= iframeRect.x;
        if (sr.x < 0) {
            dr.x -= NSToIntRound(sr.x * xScaleRatio);
            sr.width += sr.x;
            dr.width += NSToIntRound(sr.x * xScaleRatio);
            if (sr.width <= 0 || dr.width <= 0)
            return NS_OK;
            sr.x = 0;
        } else if (sr.x > iframeRect.width) {
            return NS_OK;
        }
    }

    if (iframeRect.y > 0) {
        float yScaleRatio = float(dr.height) / sr.height;

        // adjust for offset  
        sr.y -= iframeRect.y;
        if (sr.y < 0) {
            dr.y -= NSToIntRound(sr.y * yScaleRatio);
            sr.height += sr.y;
            dr.height += NSToIntRound(sr.y * yScaleRatio);
            if (sr.height <= 0 || dr.height <= 0)
            return NS_OK;
            sr.y = 0;
        } else if (sr.y > iframeRect.height) {
            return NS_OK;
        }
    }  

    struct Image ximg;
    memset(&ximg, 0, sizeof(struct Image));

    img->LockImagePixels(PR_FALSE);
    ximg.x = dr.x;
    ximg.y = dr.y;
    ximg.w = img->GetWidth();
    ximg.h = img->GetHeight();
    ximg.destw = dr.width;
    ximg.desth = dr.height;

    ximg.bytes = img->GetBits();
    ximg.nbytes = img->GetLineStride() * img->GetHeight();
    ximg.bpp = img->GetBytesPix();

    struct Request* req = mADTM->GetRequest();
    DownSampleImage(&ximg, req->device.display);

    AddAlphaData(img, &ximg);

    xcoder_Image(mADTM->GetXCoder(), &ximg);

    if (ximg.bytes != img->GetBits()) {
        free(ximg.bytes);
        ximg.bytes = NULL;
    }

    mADTM->Inc(sizeof(struct Image));
    img->UnlockImagePixels(PR_FALSE);
 
    return NS_OK;
}

NS_IMETHODIMP
anyaRenderingContext::DrawTile(imgIContainer *aImage,
                               nscoord aXOffset, nscoord aYOffset,
                               const nsRect * aTargetRect)
{
    nsRect dr(*aTargetRect);
    nsRect so(0, 0, aXOffset, aYOffset);

    mTranMatrix->TransformCoord(&dr.x, &dr.y, &dr.width, &dr.height);

    mTranMatrix->TransformCoord(&so.x, &so.y, &so.width, &so.height);

    nsCOMPtr<gfxIImageFrame> iframe;
    aImage->GetCurrentFrame(getter_AddRefs(iframe));
    if (!iframe) return NS_ERROR_FAILURE;

    nsCOMPtr<nsIImage> img(do_GetInterface(iframe));
    if (!img) return NS_ERROR_FAILURE;
    

    struct Image ximg;
    memset(&ximg, 0, sizeof(struct Image));

    img->LockImagePixels(PR_FALSE);

    ximg.istile = TRUE;
    ximg.x = dr.x;
    ximg.y = dr.y;
    ximg.w = img->GetWidth();
    ximg.h = img->GetHeight();
    ximg.destw = dr.width;
    ximg.desth = dr.height;

    ximg.bytes = img->GetBits();
    ximg.nbytes = img->GetLineStride() * img->GetHeight(); // XXXben ??? check to see if DHeight is what we should be using here. 
    ximg.bpp = img->GetBytesPix();

    struct Request* req = mADTM->GetRequest();
    DownSampleImage(&ximg, req->device.display);

    AddAlphaData(img, &ximg);

    xcoder_Image(mADTM->GetXCoder(), &ximg);

    if (ximg.bytes != img->GetBits()) {
        free(ximg.bytes);
        ximg.bytes = NULL;
    }

    mADTM->Inc(sizeof(struct Image));
    img->UnlockImagePixels(PR_FALSE);

    return NS_OK;
}

NS_IMETHODIMP
anyaRenderingContext::DrawScaledTile(imgIContainer *aImage,
                                     nscoord aXOffset, nscoord aYOffset,
                                     nscoord aTileWidth, nscoord aTileHeight,
                                     const nsRect * aTargetRect)
{
    return NS_OK;
}
